#!/usr/bin/env python3

"""
Dialog for adding a new remote.
"""

from functools import partial
import typing

from qtpy import QtCore, QtGui, QtWidgets, PYSIDE2

import validators

from cruiz.interop.pod import ConanRemote

from cruiz.settings.managers.recentconanremotes import RecentConanRemotesSettingsReader

if PYSIDE2:
    from cruiz.pyside2.local_cache_add_remote import Ui_AddRemoteDialog

    QAction = QtWidgets.QAction
else:
    from cruiz.pyside6.local_cache_add_remote import Ui_AddRemoteDialog

    QAction = QtGui.QAction


class AddRemoteDialog(QtWidgets.QDialog):
    """
    Dialog for adding a new remote.
    """

    def __init__(self, parent: QtWidgets.QWidget) -> None:
        super().__init__(parent)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose, True)
        self._ui = Ui_AddRemoteDialog()
        self._ui.setupUi(self)  # type: ignore[no-untyped-call]
        recent_remotes_menu = QtWidgets.QMenu("Recent remotes", self)
        with RecentConanRemotesSettingsReader() as settings:
            recent_remote_urls = settings.urls.resolve()
        if recent_remote_urls:
            for remote in recent_remote_urls:
                remote_action = QAction(remote, self)
                remote_action.triggered.connect(partial(self._set_remote_url, remote))
                recent_remotes_menu.addAction(remote_action)
        else:
            recent_remotes_menu.setEnabled(False)
        self._ui.url.set_custom_menu(recent_remotes_menu)
        self._ui.url.textChanged.connect(self._updated)
        self._ui.name.textChanged.connect(self._updated)
        self._ui.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).clicked.connect(
            self.accept
        )
        self._ui.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(False)
        self._ui.buttonBox.button(QtWidgets.QDialogButtonBox.Cancel).clicked.connect(
            self.reject
        )
        self._new_remote: typing.Optional[ConanRemote] = None

    def _set_remote_url(self, url: str) -> None:
        self._ui.url.setText(url)

    def _updated(self, path: str) -> None:
        # pylint: disable=unused-argument
        self._ui.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(
            bool(
                self._ui.name.text()
                and self._ui.url.text()
                and validators.url(self._ui.url.text())
            )
        )

    def accept(self) -> None:
        self._new_remote = ConanRemote(self._ui.name.text(), self._ui.url.text(), True)
        super().accept()

    @property
    def new_remote(self) -> ConanRemote:
        """
        Details of the new remote just added.
        """
        assert self._new_remote
        return self._new_remote



#!/usr/bin/env python3

"""
Dialog for adding extra profile directories
"""

import typing

from qtpy import QtCore, QtGui, QtWidgets, PYSIDE2

from cruiz.interop.pod import ExtraProfileDirectory
from cruiz.widgets.util import search_for_dir_options

if PYSIDE2:
    from cruiz.pyside2.local_cache_add_profile_directory import (
        Ui_AddExtraProfileDirectoryDialog,
    )

    QAction = QtWidgets.QAction
else:
    from cruiz.pyside6.local_cache_add_profile_directory import (
        Ui_AddExtraProfileDirectoryDialog,
    )

    QAction = QtGui.QAction


class AddExtraProfileDirectoryDialog(QtWidgets.QDialog):
    """
    Dialog for adding an extra profile directory.
    """

    def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose, True)
        self._ui = Ui_AddExtraProfileDirectoryDialog()
        self._ui.setupUi(self)  # type: ignore[no-untyped-call]
        self._ui.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(False)
        self._ui.buttonBox.accepted.connect(self.accept)
        self._ui.buttonBox.rejected.connect(self.reject)
        self._ui.name.textChanged.connect(self._updated)
        self._ui.directory.textChanged.connect(self._updated)
        browse_for_profile_dir_action = QAction(
            self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon), "", self
        )
        browse_for_profile_dir_action.triggered.connect(self._browse_for_profile_dir)
        self._ui.directory.addAction(
            browse_for_profile_dir_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        self._extra: typing.Optional[ExtraProfileDirectory] = None

    def _updated(self, text: str) -> None:
        # pylint: disable=unused-argument
        self._ui.buttonBox.button(QtWidgets.QDialogButtonBox.Ok).setEnabled(
            bool(self._ui.name.text() and self._ui.directory.text())
        )

    def accept(self) -> None:
        self._extra = ExtraProfileDirectory(
            self._ui.name.text(), self._ui.directory.text()
        )
        super().accept()

    def _browse_for_profile_dir(self) -> None:
        dir_found = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            "Select extra profile directory",
            "",
            search_for_dir_options(),
        )
        if dir_found:
            self._ui.directory.setText(dir_found)

    @property
    def extra(self) -> ExtraProfileDirectory:
        """
        The extra profile directory added.
        """
        assert self._extra
        return self._extra



#!/usr/bin/env python3

"""
Wizard for loading a recipe
"""

import pathlib
import typing

from qtpy import QtCore, QtWidgets, PYSIDE2

if PYSIDE2:
    from cruiz.pyside2.load_recipe_wizard import Ui_LoadRecipeWizard
else:
    from cruiz.pyside6.load_recipe_wizard import Ui_LoadRecipeWizard

from cruiz.exceptions import RecipeDoesNotExistError, RecipeAlreadyOpenError
from cruiz.settings.managers.recipe import RecipeSettings, RecipeSettingsReader
from cruiz.recipe.recipewidget import Recipe

from cruiz.commands.context import managed_conan_context
from cruiz.commands.logdetails import LogDetails
from cruiz.constants import DEFAULT_CACHE_NAME

import cruiz.globals


class LoadRecipeWizard(QtWidgets.QWizard):
    """
    Wizard for loading recipes.
    """

    def __init__(
        self,
        parent: QtWidgets.QWidget,
        path: pathlib.Path,
        uuid: typing.Optional[QtCore.QUuid],
        original_recipe: typing.Optional[Recipe],
    ) -> None:
        super().__init__(parent)
        self.has_load_errors = False
        self._path = path
        self._uuid = uuid
        self.original_recipe = original_recipe
        self._ambiguous_version = False
        self.validate()
        if self._uuid:
            # likely to have come via the recent recipe menu
            with RecipeSettingsReader.from_uuid(self._uuid) as settings:
                self._existing_cache_name = settings.local_cache_name.resolve()
            assert self._existing_cache_name is not None
        else:
            self.matching_uuids = RecipeSettings().matching_uuids(self._path)
            if self.matching_uuids:
                if len(self.matching_uuids) == 1:
                    recipe_uuid = self.matching_uuids[0]
                    with RecipeSettingsReader.from_uuid(recipe_uuid) as settings:
                        package_attributes = settings.attribute_overrides.resolve()
                    if "version" in package_attributes:
                        # a version has been been used as an override
                        self._ambiguous_version = True
                else:
                    self._ambiguous_version = True
        self.ui = Ui_LoadRecipeWizard()
        self.ui.setupUi(self)  # type: ignore[no-untyped-call]

        if self.disambiguation_required:
            self.ui.intro_message.hide()
            log_details = LogDetails(self.ui.intro_message, None, True, False, None)
            log_details.logging.connect(self._on_error_loading)
            with managed_conan_context(DEFAULT_CACHE_NAME, log_details) as context:
                self.recipe_attributes = context.inspect_recipe(self._path)
                self.conandata = context.get_conandata(self._path)

    def _on_error_loading(self) -> None:
        self.ui.intro_message.show()
        self.has_load_errors = True

    @property
    def disambiguation_required(self) -> bool:
        """
        Determine whether the recipe needs to be disambiguated by user input.
        """
        return self._uuid is None or self._ambiguous_version

    @property
    def recipe_version(self) -> typing.Optional[str]:
        """
        Get the version of the recipe. May be None.
        """
        if self.ui.version.isEnabled():
            return self.field("version")
        return None

    @property
    def local_cache(self) -> str:
        """
        Get the local cache to use, either from a clone of an existing recipe
        or from user selection.
        """
        if self._uuid:
            # come via the recent recipe menu, or chosen as a version
            # from 'open another version'
            with RecipeSettingsReader.from_uuid(self._uuid) as settings:
                existing_cache_name = settings.local_cache_name.resolve()
            assert existing_cache_name is not None
            return existing_cache_name
        return self.field("local_cache")

    @property
    def initial_profile(self) -> str:
        """
        Get the initial profile selected.
        """
        return self.field("initial_profile")

    @property
    def uuid(self) -> typing.Optional[QtCore.QUuid]:
        """
        Get the uuid to use for the recipe.
        May be None to indicate that the recipe has not been seen before.
        """
        return self._uuid

    @uuid.setter
    def uuid(self, uuid: typing.Optional[QtCore.QUuid]) -> None:
        """
        Allow setting the uuid. May be None.
        """
        self._uuid = uuid

    def validate(self) -> None:
        """
        Validate the recipe in the wizard.
        """
        if not self._path.exists():
            raise RecipeDoesNotExistError()
        if self._uuid:
            if cruiz.globals.get_main_window().is_recipe_active(self._uuid):
                raise RecipeAlreadyOpenError()



#!/usr/bin/env python3

"""
Interop between Conan and cruiz
"""

from __future__ import annotations

from dataclasses import dataclass
import typing

from .packagenode import PackageNode


@dataclass
class DependencyGraph:
    """
    Representation of Conan's dependency graph
    """

    nodes: typing.List[PackageNode]
    root: PackageNode


def dependencygraph_from_node_dependees(node: PackageNode) -> DependencyGraph:
    """
    Generate a DependencyGraph from the immediate parents (those nodes requiring
    the node specified) as an inverted dependency graph.
    """
    new_root = node.clone_standalone()
    for parent in node.parents:
        new_root.children.append(parent.clone_standalone())
    new_graph = DependencyGraph(
        [new_root] + new_root.children,
        new_root,
    )
    return new_graph



#!/usr/bin/env python3

"""
The Qt main window of the application
"""

from __future__ import annotations

from functools import partial
import logging
import os
import pathlib
import platform
import subprocess
import time
import typing

from qtpy import QtCore, QtGui, QtWidgets, PYSIDE6
import psutil

from cruiz.exceptions import (
    InconsistentSettingsError,
    RecipeAlreadyOpenError,
    RecipeDoesNotExistError,
    RecipeInspectionError,
)

from cruiz.recipe.recipewidget import RecipeWidget
from cruiz.settings.managers.recipe import (
    RecipeSettings,
    RecipeSettingsReader,
    RecipeSettingsWriter,
)
from cruiz.settings.preferencesdialog import PreferencesDialog
from cruiz.settings.managers.generalpreferences import GeneralSettingsReader
from cruiz.settings.managers.recentrecipes import (
    RecentRecipeSettingsReader,
    RecentRecipeSettingsWriter,
    RecentRecipeSettingsDeleter,
)
from cruiz.settings.managers.cmakepreferences import CMakeSettingsReader
from cruiz.settings.managers.ninjapreferences import NinjaSettingsReader
from cruiz.settings.managers.compilercachepreferences import CompilerCacheSettingsReader
from cruiz.settings.managers.recipe import RecipeSettingsDeleter
from cruiz.widgets import (
    AboutDialog,
    log_created_widget,
)
from cruiz.manage_local_cache import ManageLocalCachesDialog

from cruiz.environ import EnvironSaver
from cruiz.remote_browser.remotebrowser import RemoteBrowserDock

from cruiz.load_recipe.loadrecipewizard import LoadRecipeWizard

import cruiz.config

import cruiz.globals

logger = logging.getLogger(__name__)

if PYSIDE6:
    QAction = QtGui.QAction
else:
    QAction = QtWidgets.QAction


class MainWindow(QtWidgets.QMainWindow):
    """
    The main window of the application
    """

    # pylint: disable=too-many-instance-attributes, too-many-locals

    profile_dirs_changed = QtCore.Signal()
    remote_added_to_cache = QtCore.Signal(str)
    preferences_updated = QtCore.Signal()
    local_cache_changed = QtCore.Signal(str)

    def __del__(self) -> None:
        logger.debug("-=%d", id(self))

    def __init__(self) -> None:
        # pylint: disable=too-many-statements, global-statement
        super().__init__()
        log_created_widget(self, logger)

        assert not cruiz.globals.CRUIZ_MAINWINDOW

        self._systray = None
        if QtWidgets.QSystemTrayIcon.isSystemTrayAvailable():
            self._app_icon = QtGui.QIcon(":/cruiz.png")
            self._app_menu = QtWidgets.QMenu()
            self._systray = QtWidgets.QSystemTrayIcon(self._app_icon, self)
            self._systray.setContextMenu(self._app_menu)
            self._systray.show()

        mdi_area = QtWidgets.QMdiArea()
        mdi_area.setViewMode(QtWidgets.QMdiArea.TabbedView)
        mdi_area.setDocumentMode(True)
        mdi_area.setTabsClosable(True)
        mdi_area.setTabsMovable(True)
        self.setCentralWidget(mdi_area)

        self._remote_browser_dock = RemoteBrowserDock(self)
        self.local_cache_changed.connect(
            self._remote_browser_dock.on_local_cache_modified
        )
        self.addDockWidget(
            QtCore.Qt.RightDockWidgetArea,
            self._remote_browser_dock,
        )

        open_recipe_action = QAction("Open recipe...", self)
        open_recipe_action.setShortcut(QtGui.QKeySequence("Ctrl+O"))
        open_recipe_action.triggered.connect(self._open_recipe)

        self._recent_recipe_menu = QtWidgets.QMenu("Recent recipes", self)

        exit_action = QAction("&Quit", self)
        exit_action.setMenuRole(QAction.QuitRole)
        exit_action.setShortcut(QtGui.QKeySequence("Ctrl+Q"))
        exit_action.setStatusTip("Quit application")
        exit_action.triggered.connect(
            qApp.closeAllWindows,  # type: ignore # noqa: F821
            QtCore.Qt.QueuedConnection,
        )

        # need a clone of the exit action without the menu role to
        # appear on the systrayicon
        systray_exit_icon = QAction("&Quit", self)
        systray_exit_icon.setShortcut(QtGui.QKeySequence("Ctrl+Q"))
        systray_exit_icon.setStatusTip("Quit application")
        systray_exit_icon.triggered.connect(
            qApp.closeAllWindows,  # type: ignore # noqa: F821
            QtCore.Qt.QueuedConnection,
        )
        if QtWidgets.QSystemTrayIcon.isSystemTrayAvailable():
            self._app_menu.addAction(systray_exit_icon)

        menubar = self.menuBar()
        file_menu = menubar.addMenu("&File")
        file_menu.aboutToShow.connect(self._rebuild_recent_recipe_menu)
        file_menu.addAction(open_recipe_action)
        file_menu.addMenu(self._recent_recipe_menu)
        file_menu.addSeparator()
        file_menu.addAction(exit_action)

        edit_menu = menubar.addMenu("&Edit")
        edit_preferences_action = QAction("Preferences...", self)
        edit_preferences_action.setMenuRole(QAction.PreferencesRole)
        edit_preferences_action.triggered.connect(self._edit_preferences_new)
        edit_menu.addAction(edit_preferences_action)
        manage_local_caches_action = QAction("Manage local caches...", self)
        manage_local_caches_action.triggered.connect(self._manage_local_caches)
        edit_menu.addAction(manage_local_caches_action)

        view_menu = menubar.addMenu("&View")
        view_menu.addAction(self._remote_browser_dock.toggleViewAction())

        help_menu = menubar.addMenu("&Help")
        about_cruiz_action = QAction("About cruiz...", self)
        about_cruiz_action.setMenuRole(QAction.AboutRole)
        about_cruiz_action.triggered.connect(self._about_cruiz)
        help_menu.addAction(about_cruiz_action)
        about_qt_action = QAction("About Qt...", self)
        about_qt_action.setMenuRole(QAction.AboutQtRole)
        about_qt_action.triggered.connect(QtWidgets.QApplication.aboutQt)
        help_menu.addAction(about_qt_action)
        icon_license_action = QAction("Icon licenses...", self)
        icon_license_action.triggered.connect(self._icon_license)
        help_menu.addAction(icon_license_action)

        self._ccache_label: QtWidgets.QLabel = QtWidgets.QLabel()
        self._ccache_label.setFrameStyle(
            QtWidgets.QFrame.Panel | QtWidgets.QFrame.Sunken
        )
        self.statusBar().addPermanentWidget(self._ccache_label)

        self._sccache_label: QtWidgets.QLabel = QtWidgets.QLabel()
        self._sccache_label.setFrameStyle(
            QtWidgets.QFrame.Panel | QtWidgets.QFrame.Sunken
        )
        self.statusBar().addPermanentWidget(self._sccache_label)

        self._buildcache_label: QtWidgets.QLabel = QtWidgets.QLabel()
        self._buildcache_label.setFrameStyle(
            QtWidgets.QFrame.Panel | QtWidgets.QFrame.Sunken
        )
        self.statusBar().addPermanentWidget(self._buildcache_label)

        self._ssh_agent_label: QtWidgets.QLabel = QtWidgets.QLabel()
        self._ssh_agent_label.setFrameStyle(
            QtWidgets.QFrame.Panel | QtWidgets.QFrame.Sunken
        )
        self.statusBar().addPermanentWidget(self._ssh_agent_label)
        self._configure_ssh_agent_statusbar()

        self._compiler_label: QtWidgets.QLabel = QtWidgets.QLabel()
        self._compiler_label.setFrameStyle(
            QtWidgets.QFrame.Panel | QtWidgets.QFrame.Sunken
        )
        self.statusBar().addPermanentWidget(self._compiler_label)

        self._cmake_label: QtWidgets.QLabel = QtWidgets.QLabel()
        self._cmake_label.setFrameStyle(
            QtWidgets.QFrame.Panel | QtWidgets.QFrame.Sunken
        )
        self.statusBar().addPermanentWidget(self._cmake_label)

        self._ninja_label: QtWidgets.QLabel = QtWidgets.QLabel()
        self._ninja_label.setFrameStyle(
            QtWidgets.QFrame.Panel | QtWidgets.QFrame.Sunken
        )
        self.statusBar().addPermanentWidget(self._ninja_label)

        self._conan_label: QtWidgets.QLabel = QtWidgets.QLabel()
        self._conan_label.setFrameStyle(
            QtWidgets.QFrame.Panel | QtWidgets.QFrame.Sunken
        )
        self.statusBar().addPermanentWidget(self._conan_label)

        self._warning_label = QtWidgets.QLabel()
        self._warning_label.setPixmap(
            self.style()
            .standardIcon(QtWidgets.QStyle.SP_MessageBoxWarning)
            .pixmap(32, 32)
        )
        self.statusBar().addPermanentWidget(self._warning_label)
        self._warning_label.hide()
        self._warning_label.setContextMenuPolicy(QtGui.Qt.CustomContextMenu)
        self._warning_label.customContextMenuRequested.connect(
            self._on_warning_label_menu
        )

        self._status_file_watcher: QtCore.QFileSystemWatcher = (
            QtCore.QFileSystemWatcher()
        )

        self._refresh_status_bar()

        # deferred until the next event loop, to take into account files being
        # removed and readded by package managers like pip
        self._status_file_watcher.fileChanged.connect(
            self._refresh_status_bar,
            QtCore.Qt.QueuedConnection,
        )

        self.setWindowTitle(qApp.applicationName())  # type: ignore # noqa: F821

        cruiz.globals.CRUIZ_MAINWINDOW = self

    def _on_warning_label_menu(self, position: QtCore.QPoint) -> None:
        action = QAction("Dismiss", self)
        action.triggered.connect(self._on_warning_label_dismiss)
        menu = QtWidgets.QMenu(self)
        menu.addAction(action)
        menu.exec_(self.sender().mapToGlobal(position))

    def _on_warning_label_dismiss(self) -> None:
        self._warning_label.setToolTip("")
        self._warning_label.hide()

    def _configure_ssh_agent_statusbar(self) -> None:
        ssh_agent_pid = None
        ssh_agent_text = "No SSH agent"
        ssh_agent_color = "red"
        # find the ssh-agent if it is running
        for process in psutil.process_iter():
            if "ssh-agent" in process.name():
                ssh_agent_pid = process.pid
                ssh_agent_text = "SSH agent detected"
                ssh_agent_color = "darkorange"
                break
        if ssh_agent_pid:
            try:
                ssh_add_list_capture = subprocess.run(
                    ["ssh-add", "-l"],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    check=True,
                    timeout=3,
                )
                identities = ssh_add_list_capture.stdout.decode("utf-8").splitlines()
                tooltip = f"SSH agent PID {ssh_agent_pid}\n" + " ".join(identities)
                ssh_agent_color = "green"
            except subprocess.CalledProcessError as exc:
                error_text = exc.output.decode("utf-8").strip()
                tooltip = (
                    f"SSH agent PID {ssh_agent_pid}\n"
                    f"Cannot determine identities: {error_text}"
                )
            except subprocess.TimeoutExpired as exc:
                error_text = exc.output.decode("utf-8").strip() if exc.output else ""
                tooltip = (
                    f"SSH agent PID {ssh_agent_pid}\n"
                    f"Timed out communicating with ssh-agent: {error_text}"
                )
                ssh_agent_color = "red"
            self._ssh_agent_label.setToolTip(tooltip)
        self._ssh_agent_label.setText(ssh_agent_text)
        self._ssh_agent_label.setStyleSheet(f"QLabel {{color:{ssh_agent_color};}}")

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        self.centralWidget().closeAllSubWindows()
        if self.centralWidget().subWindowList():
            # if any recipe widget has blocked close, do not continue
            event.ignore()
            return
        self.removeDockWidget(self._remote_browser_dock)
        self._remote_browser_dock.cleanup()
        cruiz.globals.CRUIZ_MAINWINDOW = None
        super().closeEvent(event)

    def _indicate_recipe_settings_issue(self) -> None:
        self._warning_label.show()
        self._warning_label.setToolTip(
            "Issue found with recipe settings. Please clean via the preferences"
        )

    def _rebuild_recent_recipe_menu(self) -> None:
        self._recent_recipe_menu.clear()
        with RecentRecipeSettingsReader() as recent_reader:
            uuids = recent_reader.uuids()
        has_recipe_settings_issue = False
        all_uuids_with_settings = RecipeSettings().all_uuids
        for uuid in uuids:
            with RecipeSettingsReader.from_uuid(uuid) as settings:
                path = settings.path.resolve()
                if path is None:
                    has_recipe_settings_issue = True
                    continue
                recipe_path = pathlib.Path(path)
                if not recipe_path.exists():
                    has_recipe_settings_issue = True
                    # can drop through even though the path doesn't exist
                attributes = settings.attribute_overrides.resolve()
            label = (
                f"{recipe_path} [v{attributes['version']}]"
                if "version" in attributes
                else str(recipe_path)
            )
            recent_recipe_action = QAction(label, self)
            recent_recipe_action.triggered.connect(
                partial(self.load_recipe, recipe_path, uuid)
            )
            self._recent_recipe_menu.addAction(recent_recipe_action)
            recent_recipe_action.setEnabled(not self.is_recipe_active(uuid))
            all_uuids_with_settings.remove(uuid)
        self._recent_recipe_menu.setEnabled(len(uuids) > 0)
        if all_uuids_with_settings:
            has_recipe_settings_issue = True
        if has_recipe_settings_issue:
            self._indicate_recipe_settings_issue()

    def load_recipe(
        self,
        recipe_path: pathlib.Path,
        recipe_uuid: typing.Optional[QtCore.QUuid] = None,
        original_recipe: typing.Optional[RecipeWidget] = None,
    ) -> None:
        # pylint: disable=too-many-branches, too-many-statements
        """
        Load the Conan recipe into a usable environment
        """

        try:
            wizard = LoadRecipeWizard(
                self,
                recipe_path,
                recipe_uuid,
                original_recipe.recipe if original_recipe else None,
            )
            if wizard.disambiguation_required:
                if wizard.exec_() == QtWidgets.QDialog.Accepted:
                    if wizard.has_load_errors:
                        return
                    recipe_uuid = wizard.uuid
                    if recipe_uuid:
                        wizard.validate()
                else:
                    return
        except RecipeDoesNotExistError:
            result = QtWidgets.QMessageBox.question(
                self,
                "Recipe does not exist",
                f"The recipe at {str(recipe_path)} no longer exists. \
Remove from the recent list?",
            )
            if result == QtWidgets.QMessageBox.StandardButton.Yes:
                # TODO: this is not ideal, as it's not atomic
                # there are two QSettings created and destroyed
                # so there is a possibility that one part can fail,
                # so the settings can be left in a inconsistent state
                assert recipe_uuid
                RecentRecipeSettingsDeleter().delete(recipe_uuid)
                RecipeSettingsDeleter().delete(recipe_uuid)
            return
        except RecipeAlreadyOpenError:
            QtWidgets.QMessageBox.information(
                self,
                "Recipe already open",
                f"The recipe at {str(recipe_path)} is already open",
            )
            return
        except InconsistentSettingsError:
            self._indicate_recipe_settings_issue()
            return

        recipe_uuid = recipe_uuid or QtCore.QUuid.createUuid()

        # TODO: would be better to write both UUID and recipe details in one scope
        exists = RecentRecipeSettingsWriter().make_current(recipe_uuid)
        if not exists:
            settings = RecipeSettings()
            settings.path = str(recipe_path)  # type: ignore
            settings.local_cache_name = wizard.local_cache  # type: ignore
            settings.profile = wizard.initial_profile  # type: ignore
            if wizard.recipe_version:
                settings.append_attribute(
                    {"version": wizard.recipe_version}  # type: ignore
                )
            RecipeSettingsWriter.from_uuid(recipe_uuid).sync(settings)

        recipe_widget = RecipeWidget(recipe_path, recipe_uuid, wizard.local_cache)
        self.local_cache_changed.connect(recipe_widget.on_local_cache_changed)
        self.preferences_updated.connect(recipe_widget.on_preferences_update)
        self.centralWidget().addSubWindow(recipe_widget)
        recipe_widget.show()
        recipe_widget.setFocus(QtCore.Qt.ActiveWindowFocusReason)
        try:
            recipe_widget.post_init()
        except RecipeInspectionError:
            # this error has already been posted to the log window
            recipe_widget.failed_to_load()

    def _open_recipe(self) -> None:
        # TODO: what to do when the directory doesn't exist?
        with GeneralSettingsReader() as settings:
            default_recipe_dir = (
                settings.default_recipe_directory.resolve() or QtCore.QDir.homePath()
            )
        conan_file_path, _ = QtWidgets.QFileDialog.getOpenFileName(
            self,
            caption="Load Conan package recipe",
            dir=default_recipe_dir,
            filter="Conan Recipe (*.py *.txt)",
        )
        if not conan_file_path:
            return
        self.load_recipe(pathlib.Path(conan_file_path))

    def _edit_preferences_new(self) -> None:
        dialog = PreferencesDialog()
        dialog.preferences_updated.connect(self.preferences_updated)
        dialog.exec_()

    def _manage_local_caches(self, cache_name: typing.Optional[str] = None) -> None:
        dialog = ManageLocalCachesDialog(self, cache_name)
        dialog.cache_changed.connect(self.local_cache_changed)
        dialog.exec_()

    def _refresh(self) -> None:
        self._refresh_status_bar()

    def _refresh_status_bar(self) -> None:
        # pylint: disable=too-many-branches, too-many-statements

        def run_external_command(
            args: typing.List[str], remap_stderr: bool = False
        ) -> bytes:
            if platform.system() == "Windows":
                flag = subprocess.CREATE_NO_WINDOW  # type: ignore[attr-defined]
                output = subprocess.check_output(
                    args,
                    stderr=subprocess.STDOUT if remap_stderr else None,
                    creationflags=flag,
                )
            else:
                output = subprocess.check_output(
                    args, stderr=subprocess.STDOUT if remap_stderr else None
                )
            return output

        # TODO: this is truly awful, but give a package manager a chance to delete and
        # rewrite upgrades to packages, otherwise this code might be run when at the
        # time when files don't exist temporarily
        time.sleep(1)

        system = platform.system()
        compiler_path = None
        compiler_version_query_args = []
        if system == "Windows":
            compiler_path = QtCore.QStandardPaths.findExecutable("cl.exe")
        elif system == "Linux":
            compiler_path = QtCore.QStandardPaths.findExecutable("gcc")
            compiler_version_query_args = ["--version"]
        elif system == "Darwin":
            compiler_path = QtCore.QStandardPaths.findExecutable("clang")
            compiler_version_query_args = ["--version"]
        if compiler_path:
            self._status_file_watcher.addPath(compiler_path)
            out = run_external_command(
                [compiler_path] + compiler_version_query_args, remap_stderr=True
            )
            if out:
                compiler_version = out.decode("utf-8").splitlines()[0]
                self._compiler_label.setText(compiler_version)
            else:
                self._compiler_label.setText("Compiler detected, but no version")
            self._compiler_label.setToolTip(compiler_path)
        else:
            self._compiler_label.setText("No compiler detected")

        with EnvironSaver():
            with CMakeSettingsReader() as settings_cmake:
                cmake_dir = settings_cmake.bin_directory.resolve()
            if cmake_dir:
                os.environ["PATH"] = (
                    QtCore.QDir.toNativeSeparators(cmake_dir)
                    + os.pathsep
                    + os.environ.get("PATH", "")
                )
            cmake_exe_path = QtCore.QStandardPaths.findExecutable("cmake")
            if cmake_exe_path:
                self._status_file_watcher.addPath(cmake_exe_path)
                version_output = run_external_command(
                    [cmake_exe_path, "--version"]
                ).decode("utf-8")
                version = version_output.splitlines()[0].rsplit(" ", 1)[-1]
                self._cmake_label.setText(f"CMake v{version}")
                self._cmake_label.setToolTip(cmake_exe_path)
                self._cmake_label.setVisible(True)
            else:
                self._cmake_label.setVisible(False)

            with NinjaSettingsReader() as settings_ninja:
                ninja_dir = settings_ninja.bin_directory.resolve()
            if ninja_dir:
                os.environ["PATH"] = (
                    QtCore.QDir.toNativeSeparators(ninja_dir)
                    + os.pathsep
                    + os.environ.get("PATH", "")
                )
            ninja_exe_path = QtCore.QStandardPaths.findExecutable("ninja")
            if ninja_exe_path:
                self._status_file_watcher.addPath(ninja_exe_path)
                version_output = run_external_command(
                    [ninja_exe_path, "--version"]
                ).decode("utf-8")
                version = version_output.splitlines()[0]
                self._ninja_label.setText(f"Ninja v{version}")
                self._ninja_label.setToolTip(ninja_exe_path)
                self._ninja_label.setVisible(True)
            else:
                self._ninja_label.setVisible(False)

            with CompilerCacheSettingsReader() as settings_compilercache:
                ccache_dir = settings_compilercache.ccache_bin_directory.resolve()
                sccache_dir = settings_compilercache.sccache_bin_directory.resolve()
                buildcache_dir = (
                    settings_compilercache.buildcache_bin_directory.resolve()
                )

            if ccache_dir:
                os.environ["PATH"] = (
                    QtCore.QDir.toNativeSeparators(ccache_dir)
                    + os.pathsep
                    + os.environ.get("PATH", "")
                )
            ccache_exe_path = QtCore.QStandardPaths.findExecutable("ccache")
            if ccache_exe_path:
                self._status_file_watcher.addPath(ccache_exe_path)
                version_output = run_external_command(
                    [ccache_exe_path, "--version"]
                ).decode("utf-8")
                version = version_output.splitlines()[0]
                self._ccache_label.setText(version)
                self._ccache_label.setToolTip(ccache_exe_path)
                self._ccache_label.setVisible(True)
            else:
                self._ccache_label.setVisible(False)

            if sccache_dir:
                os.environ["PATH"] = (
                    QtCore.QDir.toNativeSeparators(sccache_dir)
                    + os.pathsep
                    + os.environ.get("PATH", "")
                )
            sccache_exe_path = QtCore.QStandardPaths.findExecutable("sccache")
            if sccache_exe_path:
                self._status_file_watcher.addPath(sccache_exe_path)
                version_output = run_external_command(
                    [sccache_exe_path, "--version"]
                ).decode("utf-8")
                version = version_output.splitlines()[0]
                self._sccache_label.setText(version)
                self._sccache_label.setToolTip(sccache_exe_path)
                self._sccache_label.setVisible(True)
            else:
                self._sccache_label.setVisible(False)

            if buildcache_dir:
                os.environ["PATH"] = (
                    QtCore.QDir.toNativeSeparators(buildcache_dir)
                    + os.pathsep
                    + os.environ.get("PATH", "")
                )
            buildcache_exe_path = QtCore.QStandardPaths.findExecutable("buildcache")
            if buildcache_exe_path:
                self._status_file_watcher.addPath(buildcache_exe_path)
                try:
                    version_output = run_external_command(
                        [buildcache_exe_path, "--version"]
                    ).decode("utf-8")
                    version = version_output.splitlines()[0]
                    self._buildcache_label.setText(version)
                    self._buildcache_label.setToolTip(buildcache_exe_path)
                    self._buildcache_label.setVisible(True)
                except subprocess.CalledProcessError:
                    self._buildcache_label.setVisible(False)
            else:
                self._buildcache_label.setVisible(False)

        conan_path = QtCore.QStandardPaths.findExecutable("conan")
        if conan_path:
            self._status_file_watcher.addPath(conan_path)
            try:
                # can fail the first time, if a migration of settings in the
                # local cache fails
                conan_version = run_external_command(
                    [conan_path, "--version"], remap_stderr=True
                )
            except subprocess.CalledProcessError:
                # second time is after the migration, so will work
                conan_version = run_external_command(
                    [conan_path, "--version"], remap_stderr=True
                )
            version_split = conan_version.decode("utf-8").splitlines()
            version_strings = [x for x in version_split if "version" in x]
            assert version_strings
            assert len(version_strings) == 1
            self._conan_label.setText(version_strings[0])
            self._conan_label.setToolTip(conan_path)

    def _about_cruiz(self) -> None:
        AboutDialog(self).exec_()

    def _icon_license(self) -> None:
        license_files = [
            "iconlicense.pdf",
            "iconlicense_cruise.pdf",
        ]
        for license in license_files:
            temp_dir = QtCore.QDir(
                QtCore.QStandardPaths.writableLocation(
                    QtCore.QStandardPaths.TempLocation
                )
            )
            temp_path = temp_dir.absoluteFilePath(license)
            if QtCore.QFile.exists(temp_path):
                # file may be read-only when copied from resources,
                # so make writeable and remove it
                perms = QtCore.QFile.permissions(temp_path)
                if perms & QtCore.QFileDevice.WriteOwner == 0:
                    perms = perms | QtCore.QFileDevice.WriteOwner
                    QtCore.QFile.setPermissions(temp_path, perms)
                QtCore.QFile.remove(temp_path)
            if not QtCore.QFile.copy(f":/{license}", temp_path):
                QtWidgets.QMessageBox.critical(
                    self,
                    "Failed file copy",
                    f"Unable to copy the {license} to a temporary location for viewing",
                )
                return
            QtGui.QDesktopServices.openUrl(
                QtCore.QUrl(f"file:///{temp_path}", QtCore.QUrl.TolerantMode)
            )

    def is_recipe_active(self, uuid: QtCore.QUuid) -> bool:
        """
        Is the recipe with the given UUID active (i.e. has an open tab)?
        """
        mdi_area = self.centralWidget()
        for subwindow in mdi_area.subWindowList():
            if subwindow.widget().recipe.uuid == uuid:
                return True
        return False



#!/usr/bin/env python3

"""
Remote browser page
"""

import typing

from qtpy import QtCore, QtGui, QtWidgets, PYSIDE2

if PYSIDE2:
    QAction = QtWidgets.QAction
else:
    QAction = QtGui.QAction

from cruiz.interop.packageidparameters import PackageIdParameters
from cruiz.widgets.util import BlockSignals

from .page import Page


class _PackageIdModel(QtCore.QAbstractTableModel):
    def __init__(self) -> None:
        super().__init__()
        self._pids: typing.Optional[
            typing.List[
                typing.Dict[
                    str,
                    typing.Union[str, bool, typing.Dict[str, str], typing.List[str]],
                ]
            ]
        ] = None
        self._options: typing.Optional[typing.List[str]] = None
        self._settings: typing.Optional[typing.List[str]] = None
        self._options_values: typing.Optional[typing.Dict[str, typing.Set[str]]] = None
        self._settings_values: typing.Optional[typing.Dict[str, typing.Set[str]]] = None

    def set(
        self,
        results: typing.Optional[
            typing.List[
                typing.Dict[
                    str,
                    typing.Union[str, bool, typing.Dict[str, str], typing.List[str]],
                ]
            ]
        ],
    ) -> None:
        """
        Set the package ids into the model
        """
        self.beginResetModel()
        self._pids = results
        if results is not None:
            options = set()
            option_values: typing.Dict[str, typing.Set[str]] = {}
            settings = set()
            settings_values: typing.Dict[str, typing.Set[str]] = {}
            for pid in results:
                if "options" in pid:
                    assert isinstance(pid["options"], dict)
                    for option, value in pid["options"].items():
                        options.add(option)
                        if option not in option_values:
                            option_values[option] = set()
                        option_values[option].add(value)
                if "settings" in pid:
                    assert isinstance(pid["settings"], dict)
                    for setting, value in pid["settings"].items():
                        settings.add(setting)
                        if setting not in settings_values:
                            settings_values[setting] = set()
                        settings_values[setting].add(value)
            self._options = sorted(list(options))
            self._settings = sorted(list(settings))
            self._options_values = option_values
            self._settings_values = settings_values
        else:
            self._options = None
            self._settings = None
            self._options_values = None
            self._settings_values = None
        self.endResetModel()

    def rowCount(self, parent) -> int:  # type: ignore
        if parent.isValid():
            return 0
        if self._pids is None:
            return 0
        return len(self._pids)

    def columnCount(self, parent) -> int:  # type: ignore
        # pylint: disable=unused-argument
        if self._pids is None:
            return 0
        num_cols = 1  # the package id
        if self._settings:
            num_cols += len(self._settings)  # type: ignore[arg-type]
        if self._options:
            num_cols += len(self._options)  # type: ignore[arg-type]
        if "requires" in self._pids[0]:
            num_cols += 1
        return num_cols

    def headerData(self, section, orientation, role):  # type: ignore
        if role == QtCore.Qt.DisplayRole and orientation == QtCore.Qt.Horizontal:
            num_settings = len(self._settings) if self._settings else 0
            num_options = len(self._options) if self._options else 0
            if section == 0:
                return "Package Id"
            if section >= 1 and section < 1 + num_settings:
                return self._settings[section - 1]  # type: ignore[index]
            if section >= 1 + num_settings and section < 1 + num_settings + num_options:
                index = section - 1 - num_settings
                return self._options[index]  # type: ignore[index]
            return "Requires"
        return None

    def data(self, index, role):  # type: ignore
        if role == QtCore.Qt.DisplayRole:
            num_settings = len(self._settings) if self._settings else 0
            num_options = len(self._options) if self._options else 0
            section = index.column()
            assert self._pids
            if section == 0:
                return self._pids[index.row()]["id"]
            if section >= 1 and section < 1 + num_settings:
                header = self.headerData(  # type: ignore[no-untyped-call]
                    section, QtCore.Qt.Horizontal, QtCore.Qt.DisplayRole
                )
                try:
                    return self._pids[index.row()]["settings"][header]
                except KeyError:
                    return "-"
            if section >= 1 + num_settings and section < 1 + num_settings + num_options:
                header = self.headerData(  # type: ignore[no-untyped-call]
                    section, QtCore.Qt.Horizontal, QtCore.Qt.DisplayRole
                )
                try:
                    return self._pids[index.row()]["options"][header]
                except KeyError:
                    return "-"
            value = "\n".join(sorted(self._pids[index.row()]["requires"]))
            return value
        return None


class _FilteringModel(QtCore.QAbstractTableModel):
    def __init__(self) -> None:
        super().__init__()
        self._filters: typing.List[typing.Tuple[int, str, str]] = []

    def add(self, index: int, key: str, value: str) -> None:
        self.beginResetModel()
        self._filters.append((index, key, value))
        self.endResetModel()

    def remove(self, index: int) -> None:
        self.beginResetModel()
        del self._filters[index]
        self.endResetModel()

    def invalidate(self) -> None:
        self.beginResetModel()
        self._filters = []
        self.endResetModel()

    def rowCount(self, parent) -> int:  # type: ignore
        return len(self._filters)

    def columnCount(self, parent) -> int:  # type: ignore
        return 2

    def data(self, index, role):  # type: ignore
        if role == QtCore.Qt.DisplayRole:
            return self._filters[index.row()][index.column() + 1]
        return None

    def headerData(self, section, orientation, role):  # type: ignore
        if role == QtCore.Qt.DisplayRole and orientation == QtCore.Qt.Horizontal:
            if 0 == section:
                return "Key"
            return "Value"
        return None


class _PackageIdSortFilterProxyModel(QtCore.QSortFilterProxyModel):
    def __init__(self, model: _PackageIdModel, filtering: _FilteringModel) -> None:
        super().__init__()
        self.setSourceModel(model)
        self._filtering = filtering

    def filterAcceptsRow(
        self,
        source_row: int,
        source_parent: typing.Union[QtCore.QModelIndex, QtCore.QPersistentModelIndex],
    ) -> bool:
        for column, _, value in self._filtering._filters:
            index = self.sourceModel().index(source_row, column, source_parent)
            if value != self.sourceModel().data(
                index, QtCore.Qt.DisplayRole  # type: ignore
            ):
                return False
        return super().filterAcceptsRow(source_row, source_parent)


class PackageIdPage(Page):
    """
    Remote browser page for displaying package ids
    """

    def setup(self, self_ui: typing.Any) -> None:
        """
        Setup the UI for the page
        """
        self._base_setup(self_ui, 2)
        self._current_pkgref: typing.Optional[str] = None
        self._model = _PackageIdModel()
        self._filtering_model = _FilteringModel()
        self._sorting_model = _PackageIdSortFilterProxyModel(
            self._model, self._filtering_model
        )
        self._ui.package_ids.setModel(self._sorting_model)
        self._ui.package_ids.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeToContents
        )
        self._ui.package_ids.horizontalHeader().setContextMenuPolicy(
            QtCore.Qt.ContextMenuPolicy.CustomContextMenu
        )
        self._ui.package_ids.horizontalHeader().customContextMenuRequested.connect(
            self._on_package_id_header_menu
        )
        self._ui.package_ids.verticalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeToContents
        )
        self._ui.package_ids.customContextMenuRequested.connect(
            self._on_selected_pkgref_menu
        )

        self._ui.package_ids.doubleClicked.connect(self._on_pid_dclicked)
        self._ui.pid_back.clicked.connect(self._on_back)
        self._ui.pid_cancel.clicked.connect(self.on_cancel)
        self._ui.pid_refresh.clicked.connect(self._on_refresh)
        self._ui.pid_restart.clicked.connect(self._on_restart)
        self._ui.pid_pkgref.customContextMenuRequested.connect(self._on_pkgref_menu)
        self._ui.pid_filterTable.setModel(self._filtering_model)
        self._ui.pid_filterTable.customContextMenuRequested.connect(
            self._on_pid_filter_menu
        )
        self._ui.pid_filter_key.currentTextChanged.connect(self._on_pid_filter_changed)
        self._ui.pid_filter_value.currentTextChanged.connect(
            self._on_pid_filter_value_changed
        )
        self._ui.pid_add_filter.clicked.connect(self._on_add_pid_filter)

    @property
    def package_reference(self) -> str:
        """
        Get the package reference selected in the current page
        """
        selection = self._ui.package_ids.selectedIndexes()
        assert (
            len(selection) > 0
        )  # hard to determine the number, but it's only the first we want
        package_id = self._sorting_model.data(
            selection[0], QtCore.Qt.DisplayRole  # type: ignore
        )
        return f"{self._previous_pkgref}:{package_id}"

    def _enable_progress(self, enable: bool) -> None:
        self._ui.pid_progress.setMaximum(0 if enable else 1)
        self._ui.pid_buttons.setEnabled(not enable)
        self._ui.pid_cancel.setEnabled(enable)

    def showEvent(self, event: QtGui.QShowEvent) -> None:
        # pylint: disable=unused-argument
        self._ui.pid_pkgref.setText(self._previous_pkgref)
        self._compute()

    def _complete(self, results: typing.Any, exception: typing.Any) -> None:
        self._enable_progress(False)
        if exception:
            self._ui.pid_groupbox.setTitle("Error finding package ids")
            self._log_details.stderr(str(exception))
            return
        if results is not None:
            self._model.set(results)
            self._ui.pid_groupbox.setTitle(f"{len(results)} package ids found")
            self._ui.pid_groupbox.setEnabled(True)
            self._current_pkgref = self._previous_pkgref
            self._ui.pid_filter_group.setEnabled(True)
            self._populate_settings_options_filter()

    def _populate_settings_options_filter(self) -> None:
        self._ui.pid_filter_value.setEnabled(False)
        self._ui.pid_add_filter.setEnabled(False)
        # brute force re-add all the settings and options
        with BlockSignals(self._ui.pid_filter_key) as blocked_widget:
            blocked_widget.clear()
            blocked_widget.addItem("Package Id")
            blocked_widget.addItems(self._model._settings)
            blocked_widget.addItems(self._model._options)
            blocked_widget.setCurrentIndex(-1)
        model = self._ui.pid_filter_key.model()
        # disable those with active filters
        for index, _, _ in self._filtering_model._filters:
            item = model.item(index)
            item.setEnabled(False)
        with BlockSignals(self._ui.pid_filter_value) as blocked_widget:
            blocked_widget.clear()
            blocked_widget.setCurrentIndex(-1)

    def _on_back(self) -> None:
        if self._revisions_enabled:
            self._open_previous_page()
        else:
            # skip recipe revisions
            self.parent().setCurrentIndex(self.page_index - 2)

    def _on_pid_dclicked(self, index: QtCore.QModelIndex) -> None:
        # pylint: disable=unused-argument
        if self._revisions_enabled:
            self._open_next_page()
        else:
            # skip package revisions
            self.parent().setCurrentIndex(self.page_index + 2)

    def on_cancel(self) -> None:
        """
        Called when the user cancels the operation.
        """
        self._context.cancel()
        self._enable_progress(False)

    def _on_refresh(self) -> None:
        self._current_pkgref = None
        self._compute()

    def _compute(self) -> None:
        pkgref = self._previous_pkgref
        if self._current_pkgref != pkgref:
            self._ui.pid_groupbox.setTitle("0 package ids found")
            self._model.set(None)
            self._enable_progress(True)
            self._ui.pid_groupbox.setEnabled(False)
            self._ui.pid_filter_group.setEnabled(False)
            params = PackageIdParameters(
                reference=pkgref,
                remote_name=self._ui.remote.currentText(),
            )
            self._context.get_package_details(params, self._log_details, self._complete)

    def _on_restart(self) -> None:
        self._filtering_model.invalidate()
        self._sorting_model.invalidateFilter()
        self._open_start()

    def _on_copy_pkgref_to_clip(self) -> None:
        QtWidgets.QApplication.clipboard().setText(self._ui.pid_pkgref.text())

    def _on_package_id_header_menu(self, position: QtCore.QPoint) -> None:
        menu = QtWidgets.QMenu(self)
        offset = 1
        assert self._model._settings
        for i, s in enumerate(self._model._settings):
            action = QAction(s, self)
            action.setCheckable(True)
            action.setData(i + offset)
            action.setChecked(not self._ui.package_ids.isColumnHidden(i + offset))
            action.toggled.connect(self._on_toggle_hide_column)
            menu.addAction(action)
        menu.addSeparator()
        offset = 1 + len(self._model._settings)
        assert self._model._options
        for i, o in enumerate(self._model._options):
            action = QAction(o, self)
            action.setData(i + offset)
            action.setCheckable(True)
            action.setChecked(not self._ui.package_ids.isColumnHidden(i + offset))
            action.toggled.connect(self._on_toggle_hide_column)
            menu.addAction(action)
        menu.addSeparator()
        offset = 1 + len(self._model._settings) + len(self._model._options)
        action = QAction("Requires", self)
        action.setData(offset)
        action.setCheckable(True)
        action.setChecked(not self._ui.package_ids.isColumnHidden(offset))
        action.toggled.connect(self._on_toggle_hide_column)
        menu.addAction(action)
        menu.exec_(self.sender().mapToGlobal(position))

    def _on_toggle_hide_column(self, checked: bool) -> None:
        action = self.sender()
        self._ui.package_ids.setColumnHidden(action.data(), not checked)

    def _on_pid_filter_changed(self, text: str) -> None:
        self._ui.pid_filter_value.setEnabled(True)
        assert self._model._settings
        if 0 == self.sender().currentIndex():
            with BlockSignals(self._ui.pid_filter_value) as blocked_widget:
                blocked_widget.clear()
                assert self._model._pids
                for pid in self._model._pids:
                    blocked_widget.addItem(pid["id"])
                blocked_widget.setCurrentIndex(-1)
        elif self.sender().currentIndex() < len(self._model._settings) + 1:
            with BlockSignals(self._ui.pid_filter_value) as blocked_widget:
                blocked_widget.clear()
                assert self._model._settings_values
                blocked_widget.addItems(self._model._settings_values[text])
                blocked_widget.setCurrentIndex(-1)
        else:
            with BlockSignals(self._ui.pid_filter_value) as blocked_widget:
                blocked_widget.clear()
                assert self._model._options_values
                blocked_widget.addItems(self._model._options_values[text])
                blocked_widget.setCurrentIndex(-1)

    def _on_pid_filter_value_changed(self, text: str) -> None:
        self._ui.pid_add_filter.setEnabled(True)

    def _on_add_pid_filter(self) -> None:
        self._filtering_model.add(
            self._ui.pid_filter_key.currentIndex(),
            self._ui.pid_filter_key.currentText(),
            self._ui.pid_filter_value.currentText(),
        )
        self._sorting_model.invalidateFilter()
        self._populate_settings_options_filter()

    def _on_pid_filter_menu(self, position: QtCore.QPoint) -> None:
        if not self.sender().selectionModel().selectedRows():
            return
        menu = QtWidgets.QMenu(self)
        remove_action = QAction("Remove", self)
        remove_action.triggered.connect(self._on_pid_filter_remove)
        menu.addAction(remove_action)
        menu.exec_(self.sender().mapToGlobal(position))

    def _on_pid_filter_remove(self) -> None:
        selected_row_index = (
            self._ui.pid_filterTable.selectionModel().selectedRows()[0].row()
        )
        self._filtering_model.remove(selected_row_index)
        self._sorting_model.invalidateFilter()
        self._populate_settings_options_filter()



#!/usr/bin/env python3

"""
Wizard page for selecting the version of the package
"""

import typing

from qtpy import QtGui, QtWidgets

from cruiz.widgets.util import BlockSignals
from cruiz.settings.managers.conanpreferences import ConanSettingsReader
from cruiz.settings.managers.recipe import RecipeSettingsReader

import cruiz.globals


class LoadRecipePackageVersionPage(QtWidgets.QWizardPage):
    """
    Wizard page for selecting the recipe version to bind to.
    """

    @property
    def _ui(self) -> typing.Any:
        return self.wizard().ui

    def nextId(self) -> int:
        if self._ui.version.currentText() in self._uuid_versions:
            return -1
        return 2

    def initializePage(self) -> None:
        self.registerField("version*", self._ui.version, "currentText")  # type: ignore

        self._ui.version.currentTextChanged.connect(self._on_version_changed)

        self._uuid_versions = {}
        if self.wizard().recipe_attributes["version"] is None:
            # get versions from conandata.yml
            versions = self._get_versions_from_conandata(self.wizard().conandata)
            if versions:
                # get the versions from uuids with the same paths
                for uuid in self.wizard().matching_uuids:
                    with RecipeSettingsReader.from_uuid(uuid) as settings:
                        attributes = settings.attribute_overrides.resolve()
                    if "version" in attributes:
                        self._uuid_versions[attributes["version"]] = uuid

                with BlockSignals(self._ui.version) as blocked_widget:
                    blocked_widget.clear()
                    model = blocked_widget.model()
                    for i, version in enumerate(versions):
                        if version in self._uuid_versions:
                            blocked_widget.addItem(QtGui.QIcon(":/cruiz.png"), version)
                            if cruiz.globals.get_main_window().is_recipe_active(
                                self._uuid_versions[version]
                            ):
                                item = model.item(i)
                                item.setFlags(item.flags() & ~QtGui.Qt.ItemIsEnabled)
                        else:
                            blocked_widget.addItem(version)
                    blocked_widget.setCurrentIndex(-1)
        else:
            with BlockSignals(self._ui.version) as blocked_widget:
                blocked_widget.clear()
                blocked_widget.addItem(self.wizard().recipe_attributes["version"])
            # combobox disabled to ensure it's clear that there's no choice
            self._ui.version.setEnabled(False)
        super().initializePage()

    def cleanupPage(self) -> None:
        self._ui.version.currentTextChanged.disconnect()
        return super().cleanupPage()

    def _on_version_changed(self, text: str) -> None:
        if text in self._uuid_versions:
            self.setFinalPage(True)
            self.wizard().uuid = self._uuid_versions[text]
        else:
            self.wizard().uuid = None
        self.completeChanged.emit()

    def _get_versions_from_conandata(
        self, conandata: typing.Dict[str, typing.Dict[str, str]]
    ) -> typing.Optional[typing.List[str]]:
        if not conandata:
            QtWidgets.QMessageBox.information(
                self,
                "No conandata.yaml file beside recipe",
                "Unable to read the conandata.yml beside the recipe. "
                "Does the file exist?\n"
                "The package version number can be manually specified, "
                "or the conanfile.yml can be added next to the recipe.",
            )
            return None
        with ConanSettingsReader() as settings:
            path_segments = settings.conandata_version_yaml_pathsegment.resolve()
        if path_segments is None:
            QtWidgets.QMessageBox.information(
                self,
                "Cannot identify recipe version numbers from conandata.yml file",
                "The conandata.yml beside the recipe cannot be parsed for version "
                "numbers.\n\n"
                "This is because cruiz cannot determine its format without a hint.\n\n"
                "Complete the 'Preferences->Conan->Parsing conandata' setting to "
                "specify how to find the list of version numbers for your recipe.\n\n"
                "For example, if your conandata.yml is formatted as\n\n"
                "\tsources:\n"
                "\t  1.2.3:\n"
                "\t    ...\n"
                "\t  4.5.6:\n"
                "\t    ...\n\n"
                "then specify 'sources' in the preferences.",
            )
            return None
        current_dict = conandata
        for key in path_segments.split("/"):
            if key in current_dict:
                current_dict = conandata[key]  # type: ignore[assignment]
            else:
                QtWidgets.QMessageBox.information(
                    self,
                    "Cannot find YAML path segment in recipe conandata.yml file",
                    f"Unable to locate YAML path segment '{key}' in conandata.yml "
                    f" from preferences expression '{path_segments}'.\n\nThe version "
                    "number can be manually specified, or the preference corrected for "
                    "the YAML layout.",
                )
                return None
        return list(current_dict.keys())



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import contextlib
import multiprocessing
import os
import typing

from cruiz.interop.commandparameters import CommandParameters

from .utils import worker
from .utils.message import Message, Success


@contextlib.contextmanager
def use_package_cwd(cwd: str) -> typing.Iterator[None]:
    """
    Context manager to change directory and back again when finished.
    """
    curdir = os.getcwd()
    try:
        os.chdir(cwd)
        yield
    finally:
        os.chdir(curdir)


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run an arbitrary command
    """
    with worker.ConanWorker(queue, params) as api:
        from conans.client.command import Command

        command = Command(api)
        verb_function = getattr(command, params.verb)
        result = verb_function(params.arguments)
        queue.put(Success(result))



#!/usr/bin/env python3

"""
Conan dependency graph Qt model
"""

from qtpy import QtCore, QtGui

from cruiz.interop.dependencygraph import DependencyGraph


class DependenciesListModel(QtGui.QStandardItemModel):
    """
    Qt model representing the list of dependencies to a recipe
    """

    def __init__(self, graph: DependencyGraph) -> None:
        super().__init__(len(graph.nodes), 1)
        for row, node in enumerate(graph.nodes):
            item = QtGui.QStandardItem(node.reference)
            item.setData(node, QtCore.Qt.UserRole)  # type: ignore
            if node == graph.root:
                font = QtGui.QFont()
                font.setBold(True)
                font.setUnderline(True)
                item.setData(font, QtCore.Qt.FontRole)  # type: ignore
            else:
                if not node.is_runtime:
                    font = QtGui.QFont()
                    font.setItalic(True)
                    item.setData(font, QtCore.Qt.FontRole)  # type: ignore
            node_info = f"Conan info:\n{node.info}" if node.info else ""
            item.setData(
                f"Package reference: {node.reference}\n"
                f"Package Id: {node.package_id}\n"
                f"Recipe revision: {node.recipe_revision}\n"
                f"{node_info}",
                QtCore.Qt.ToolTipRole,  # type: ignore
            )
            self.setItem(row, 0, item)


class DependenciesTreeModel(QtGui.QStandardItemModel):
    """
    Qt model representing the tree of dependencies to a recipe
    """

    def __init__(self, graph: DependencyGraph) -> None:
        super().__init__()
        parent_item = self.invisibleRootItem()
        # TODO: revisit this algorithm
        for node in graph.nodes:
            item = QtGui.QStandardItem(node.reference)
            if node == graph.root:
                item.setData(
                    QtGui.QColor(QtCore.Qt.red),
                    QtCore.Qt.ForegroundRole,  # type: ignore
                )
            else:
                if node.is_runtime:
                    item.setData(
                        QtGui.QColor(QtCore.Qt.black),
                        QtCore.Qt.ForegroundRole,  # type: ignore
                    )
                else:
                    item.setData(
                        QtGui.QColor(QtCore.Qt.gray),
                        QtCore.Qt.ForegroundRole,  # type: ignore
                    )
            parent_item.appendRow(item)
            parent_item = item



#!/usr/bin/env python3

"""
Settings context manager for compiler cache preferences
"""

import typing

from .basesettings import (
    BaseSettings,
    StringSetting,
    ComparableCommonSettings,
    SettingMeta,
)
from .writermixin import _WriterMixin
from .valueclasses import ScalarValue


class CompilerCacheSettings(ComparableCommonSettings):
    """
    Represenation of compiler cache settings
    """

    def __init__(self) -> None:
        self._property_meta = {
            "default": SettingMeta("Default", StringSetting, "None", ScalarValue),
            "ccache_bin_directory": SettingMeta(
                "ccache/BinDir", StringSetting, None, ScalarValue
            ),
            "sccache_bin_directory": SettingMeta(
                "sccache/BinDir", StringSetting, None, ScalarValue
            ),
            "buildcache_bin_directory": SettingMeta(
                "buildcache/BinDir", StringSetting, None, ScalarValue
            ),
        }

    @property
    def default(self) -> StringSetting:
        """
        Get the default compiler cache
        """
        return self._get_value_via_meta()

    @default.setter
    def default(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def ccache_bin_directory(self) -> StringSetting:
        """
        Get the ccache bin directory
        """
        return self._get_value_via_meta()

    @ccache_bin_directory.setter
    def ccache_bin_directory(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def sccache_bin_directory(self) -> StringSetting:
        """
        Get the sccache bin directory
        """
        return self._get_value_via_meta()

    @sccache_bin_directory.setter
    def sccache_bin_directory(self, value: str) -> None:
        self._set_value_via_meta(value)

    @property
    def buildcache_bin_directory(self) -> StringSetting:
        """
        Get the buildcache bin directory
        """
        return self._get_value_via_meta()

    @buildcache_bin_directory.setter
    def buildcache_bin_directory(self, value: str) -> None:
        self._set_value_via_meta(value)


class CompilerCacheSettingsReader:
    """
    Context manager for reading compiler cache settings from disk
    """

    def __init__(self) -> None:
        self.group = "Thirdparty/CompilerCache"
        self.settings = BaseSettings.make_settings()

    def __enter__(self) -> CompilerCacheSettings:
        self.settings.beginGroup(self.group)
        self._settings_object = CompilerCacheSettings()
        self._settings_object.settings_reader = self
        return self._settings_object

    def __exit__(
        self, exc_type: typing.Any, exc_value: typing.Any, exc_traceback: typing.Any
    ) -> typing.Any:
        self.settings.endGroup()
        self._settings_object.settings_reader = None
        del self._settings_object
        if exc_type is None:
            pass
        else:
            # propagate exception
            return False


class CompilerCacheSettingsWriter(_WriterMixin):
    """
    Utility for writing changed compiler cache settings to disk
    """

    def __init__(self) -> None:
        self._reader_for_writer = CompilerCacheSettingsReader()



#!/usr/bin/env python3

"""
Wizard for creating new Conan local caches
"""

import platform
import typing

from qtpy import QtGui, QtWidgets, PYSIDE2

from cruiz.commands.context import ConanContext
from cruiz.interop.commandparameters import CommandParameters
from cruiz.commands.logdetails import LogDetails
from cruiz.settings.managers.localcachepreferences import LocalCacheSettingsReader
from cruiz.settings.managers.namedlocalcache import NamedLocalCacheCreator
from cruiz.widgets.util import search_for_dir_options
from cruiz.constants import DEFAULT_CACHE_NAME

import cruiz.workers

if PYSIDE2:
    from cruiz.pyside2.local_cache_new_wizard import Ui_NewLocalCacheWizard

    QAction = QtWidgets.QAction
else:
    from cruiz.pyside6.local_cache_new_wizard import Ui_NewLocalCacheWizard

    QAction = QtGui.QAction


class NewLocalCacheWizard(QtWidgets.QWizard):
    """
    Wizard for creating new local caches
    """

    def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent)
        self._ui = Ui_NewLocalCacheWizard()
        self._ui.setupUi(self)  # type: ignore[no-untyped-call]
        self._log_details = LogDetails(self._ui.summary_log, None, True, False, None)
        self._context = ConanContext(DEFAULT_CACHE_NAME, self._log_details)
        self._ui.new_cache_name.textChanged.connect(self._ui.namePage.completeChanged)
        self._ui.userHome.textChanged.connect(self._ui.locationsPage.completeChanged)
        self._ui.userHomeShort.textChanged.connect(
            self._ui.locationsPage.completeChanged
        )
        dir_icon: QtGui.QIcon = self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon)
        browse_home_dir_action = QAction(dir_icon, "", self)
        browse_home_dir_action.triggered.connect(self._browse_for_home_dir)
        self._ui.userHome.addAction(
            browse_home_dir_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        browse_short_home_dir_action = QAction(dir_icon, "", self)
        browse_short_home_dir_action.triggered.connect(self._browse_for_short_home_dir)
        self._ui.userHomeShort.addAction(
            browse_short_home_dir_action,
            QtWidgets.QLineEdit.TrailingPosition,
        )
        with LocalCacheSettingsReader() as settings:
            new_config_url = settings.new_configuration_install.resolve()
            new_config_branch = settings.new_configuration_git_branch.resolve()
        self._ui.configUrl.setText(new_config_url)
        self._ui.configBranch.setText(new_config_branch or "<default branch>")
        self._ui.queryConfigInstall.setChecked(bool(new_config_url))
        self._ui.queryConfigInstall.setEnabled(bool(new_config_url))
        self._ui.createCache.clicked.connect(self._create_cache)
        self._ui.createProgress.setMinimum(0)
        self._ui.createProgress.setMaximum(1)
        self._ui.createProgress.setValue(0)
        if platform.system() == "Windows":
            pass
        else:
            self._ui.userHomeShortExplanation.hide()
            self._ui.userHomeShortLabel.hide()
            self._ui.userHomeShort.hide()
        self.currentIdChanged.connect(self._changed_page)

    def done(self, result: int) -> None:
        self._context.close()
        super().done(result)

    def _changed_page(self, page_id: int) -> None:
        if page_id == 4:
            # summary page
            self._ui.summary_name.setText(self._ui.new_cache_name.text())
            self._ui.summary_user_home.setText(self._ui.userHome.text())
            if self._ui.userHomeShort.text():
                self._ui.summary_user_home_short.setText(self._ui.userHomeShort.text())
            else:
                self._ui.summary_user_home_short.hide()
            if self._ui.queryConfigInstall.isChecked():
                if self._ui.configBranch.text():
                    self._ui.summary_install_config.setText(
                        "User chose to install the configuration "
                        f"{self._ui.configUrl.text()} "
                        f"using branch {self._ui.configBranch.text()}"
                    )
                else:
                    self._ui.summary_install_config.setText(
                        "User chose to install the configuration "
                        f"{self._ui.configUrl.text()} "
                        "using the default branch"
                    )
            else:
                self._ui.summary_install_config.setText(
                    "User chose not to install a configuration at this time"
                )

    def _browse_for_home_dir(self) -> None:
        dir_found = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            "Select home directory",
            "",
            search_for_dir_options(),
        )
        if dir_found:
            self._ui.userHome.setText(dir_found)

    def _browse_for_short_home_dir(self) -> None:
        dir_found = QtWidgets.QFileDialog.getExistingDirectory(
            self,
            "Select short home directory",
            "",
            search_for_dir_options(),
        )
        if dir_found:
            self._ui.userHomeShort.setText(dir_found)

    def _create_cache(self) -> None:
        self._ui.createProgress.setMaximum(0)
        name = self._ui.new_cache_name.text().strip()
        NamedLocalCacheCreator().create(
            name,
            self._ui.userHome.text().strip(),
            self._ui.userHomeShort.text().strip(),
        )

        if self._ui.queryConfigInstall.isChecked():
            with LocalCacheSettingsReader() as settings:
                config_install = settings.new_configuration_install.resolve()
                git_branch = settings.new_configuration_git_branch.resolve()
            if config_install:
                self._context.change_cache(name)
                params = CommandParameters(
                    "install_config", cruiz.workers.configinstall.invoke
                )
                named_args = params.named_arguments
                named_args["pathOrUrl"] = config_install
                if git_branch:
                    named_args["gitBranch"] = git_branch
                self._context.install_config(
                    params, self._perform_new_cache_config_install_complete
                )
        self._ui.createProgress.setMaximum(1)
        self._ui.createPage.created = True
        self._ui.createPage.completeChanged.emit()

    def _perform_new_cache_config_install_complete(
        self, result: typing.Any, exception: typing.Any
    ) -> None:
        # pylint: disable=unused-argument
        if exception:
            QtWidgets.QMessageBox.critical(
                self,
                "New local cache configuration install failure",
                str(exception),
            )

    @property
    def switch_to_new_cache(self) -> bool:
        """
        Should the user switch to the new cache once complete?
        """
        return self._ui.summary_switch_to_new.isChecked()

    @property
    def cache_name(self) -> str:
        """
        The name of the new cache once complete
        """
        return self._ui.new_cache_name.text()



#!/usr/bin/env python3

"""
Settings context managers

*Settings can be instanced either on it's own (write only) to accumulate changes or
from a *SettingsReader context manager, in order to read from the 'current settings'
Reading is on-demand via properties.
Writing does no work up-front, sets internal attributes with property setters, but then
the *SettingsWriter performs a sync operation, which compares 'current settings' against
those populated into the *Settings internal attributes.
For large *Settings, we don't need the *SettingsWriter overhead of pre-populating all
properties which is overkill if only one property changes.
The *SettingsReader usage in *Settings instances is 99% of the time to get access to the
QSettings instance that the reader setup.
The *SettingsWriter currently duplicates the group definition with the *SettingsReader.
This should be de-duplicated.
The *SettingsWriter needs a *SettingsReader in order to compare values, so
*SettingsReader is the one place the group should be defined.
If the *SettingsReader needs a parameter to initialise it, this should be mimick'd in
the *SettingsWriter initialiser:
- FontSettingsReader, NamedLocalCacheSettingsReader, RecipeSettingsReader

"""



#!/usr/bin/env python3

"""
Wizard pages for naming new Conan local caches
"""

import platform
import typing

from qtpy import QtWidgets

from cruiz.settings.managers.namedlocalcache import AllNamedLocalCacheSettingsReader


class NewLocalCacheNamePage(QtWidgets.QWizardPage):
    """
    Wizard page method overrides for specifying the new local cache name
    """

    def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent)
        with AllNamedLocalCacheSettingsReader() as names:
            self._cache_names = names

    def isComplete(self) -> bool:
        # pylint: disable=protected-access
        text = self.wizard()._ui.new_cache_name.text()
        return bool(text) and text not in self._cache_names


class NewLocalCacheLocationPage(QtWidgets.QWizardPage):
    """
    Wizard page method overrides for specifying the new local cache locations
    """

    def isComplete(self) -> bool:
        # pylint: disable=protected-access
        wizard_user_interface = self.wizard()._ui
        if platform.system() == "Windows":
            return bool(wizard_user_interface.userHome.text()) and bool(
                wizard_user_interface.userHomeShort.text()
            )
        return bool(wizard_user_interface.userHome.text())


class NewLocalCacheCreatePage(QtWidgets.QWizardPage):
    """
    Wizard page method overrides for creating the local cache.
    """

    def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent)
        self._created = False

    @property
    def created(self) -> bool:
        """
        Was the local cache created?
        """
        return self._created

    @created.setter
    def created(self, value: bool) -> None:
        self._created = value

    def isComplete(self) -> bool:
        return self.created



#!/usr/bin/env python3

"""
Converting text to HTML
"""

import html as html_module

from io import StringIO


def text_to_html(text: str, spaces_per_tab: int = 4) -> str:
    """
    Break a block of text into text and HTML markup.
    """
    html = StringIO()
    stripped_line = text.lstrip()
    num_leading_chars = len(text) - len(stripped_line)
    for index in range(num_leading_chars):
        current_char = text[index]
        if current_char == " ":
            html.write("&nbsp;")
        elif current_char == "\t":
            html.write(spaces_per_tab * "&nbsp;")
        elif current_char in ("\r", "\n"):
            continue  # dealt with later
        else:
            raise RuntimeError(
                f"Unrecognised space character: '{current_char}' "
                f"({ord(current_char)}), in the text '{text}'"
            )
    escaped = html_module.escape(stripped_line)
    escaped = escaped.replace(" ", "&nbsp;")
    html.write(escaped)
    html_str = html.getvalue()
    html_str = html_str.replace("\n", "<br>")
    html_str = html_str.replace("\t", 4 * "&nbsp;")
    return html_str



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.reciperevisionsparameters import RecipeRevisionsParameters

from .utils import worker
from .utils.message import Message, Success


def invoke(
    queue: multiprocessing.Queue[Message], params: RecipeRevisionsParameters
) -> None:
    """
    Equivalent to

    'conan search -r <remote_name> <reference> -rev'

    RecipeRevisionsParameters has dynamic attributes.
    """
    with worker.ConanWorker(queue, params) as api:
        result = api.get_recipe_revisions(
            params.reference,  # type: ignore
            remote_name=params.remote_name,  # type: ignore
        )

        queue.put(Success(result))



#!/usr/bin/env python3

"""
Widget for the table of local cache remote
"""

from enum import IntEnum
import typing

from qtpy import QtCore, QtGui, QtWidgets

from cruiz.interop.pod import ConanRemote
from cruiz.widgets.util import BlockSignals


class RemotesTable(QtWidgets.QTableWidget):
    """
    Table of remotes that can handle moving rows
    """

    class _ColumnIndex(IntEnum):
        """
        Column indices of the remotes table
        """

        ENABLED = 0
        NAME = 1
        URL = 2

    remotes_reordered = QtCore.Signal()
    remote_enabled = QtCore.Signal(str, QtCore.Qt.CheckState)

    def __init__(self, parent: typing.Optional[QtWidgets.QWidget] = None) -> None:
        super().__init__(parent)
        self.itemChanged.connect(self._item_changed)

    def _item_changed(self, item: QtWidgets.QTableWidgetItem) -> None:
        if item.column() == RemotesTable._ColumnIndex.ENABLED:
            name_item = self.item(item.row(), RemotesTable._ColumnIndex.NAME)
            if name_item:
                self.remote_enabled.emit(name_item.text().strip(), item.checkState())

    def dropEvent(self, event: QtGui.QDropEvent) -> None:
        if self == event.source():
            index_at_drop = self.indexAt(event.pos())
            source_row = self.selectedItems()[0].row()
            source_items = []
            for col in range(self.columnCount()):
                source_items.append(self.takeItem(source_row, col))
            self.removeRow(source_row)
            self.insertRow(index_at_drop.row())
            for i, item in enumerate(source_items):
                self.setItem(index_at_drop.row(), i, item)
            self.remotes_reordered.emit()
            return
        super().dropEvent(event)

    def add_remote(self, remote: ConanRemote) -> None:
        """
        Add the specified remote to the table.
        """
        row_count = self.rowCount()
        enabled_item = QtWidgets.QTableWidgetItem()
        enabled_item.setFlags(
            QtCore.Qt.ItemIsEnabled  # type: ignore
            | QtCore.Qt.ItemIsSelectable
            | QtCore.Qt.ItemIsDragEnabled
            | QtCore.Qt.ItemIsDropEnabled
            | QtCore.Qt.ItemIsUserCheckable
        )
        enabled_item.setCheckState(
            QtCore.Qt.Checked if remote.enabled else QtCore.Qt.Unchecked
        )
        name_item = QtWidgets.QTableWidgetItem(remote.name)
        name_item.setFlags(
            QtCore.Qt.ItemIsEnabled  # type: ignore
            | QtCore.Qt.ItemIsSelectable
            | QtCore.Qt.ItemIsDragEnabled
            | QtCore.Qt.ItemIsDropEnabled
        )
        url_item = QtWidgets.QTableWidgetItem(remote.url)
        url_item.setFlags(
            QtCore.Qt.ItemIsEnabled  # type: ignore
            | QtCore.Qt.ItemIsSelectable
            | QtCore.Qt.ItemIsDragEnabled
            | QtCore.Qt.ItemIsDropEnabled
        )
        with BlockSignals(self):
            self.setRowCount(row_count + 1)
            self.setItem(row_count, RemotesTable._ColumnIndex.ENABLED, enabled_item)
            self.setItem(row_count, RemotesTable._ColumnIndex.NAME, name_item)
            self.setItem(row_count, RemotesTable._ColumnIndex.URL, url_item)

    def _row_to_remote(self, row_index: int) -> ConanRemote:
        remote = ConanRemote(
            self.item(row_index, RemotesTable._ColumnIndex.NAME).text().strip(),
            self.item(row_index, RemotesTable._ColumnIndex.URL).text().strip(),
            self.item(row_index, RemotesTable._ColumnIndex.ENABLED).checkState()
            == QtCore.Qt.Checked,
        )
        return remote

    def remotes_list(self) -> typing.List[ConanRemote]:
        """
        Return list of remotes from the table.
        """
        remotes = []
        for i in range(self.rowCount()):
            remotes.append(self._row_to_remote(i))
        return remotes

    def remove_selected(self) -> ConanRemote:
        """
        Remove the selected row from the table of remotes.

        Note: single selection only, and rows only.
        """
        selected_ranges = self.selectedRanges()
        assert selected_ranges
        row_index = selected_ranges[0].topRow()
        removed_remote = self._row_to_remote(row_index)
        assert row_index == selected_ranges[0].bottomRow()
        self.removeRow(row_index)
        return removed_remote

    def same(self, remotes: typing.List[ConanRemote]) -> bool:
        """
        Are the list of remotes passed in the same as those in the table?

        There may be more rows in the table than what is in the stored list
        so we only check those from the stored list.
        """
        assert self.rowCount() >= len(remotes)
        for i, remote in enumerate(remotes):
            table_remote_name = self.item(i, RemotesTable._ColumnIndex.NAME).text()
            if table_remote_name != remote.name:
                return False
        return True



#!/usr/bin/env python3

"""
Widget for a table of key-value pairs
"""

from enum import IntEnum

from qtpy import QtCore, QtWidgets

from cruiz.widgets.util import BlockSignals


class KeyValueTable(QtWidgets.QTableWidget):
    """
    Table of key-value pairs
    """

    class ColumnIndex(IntEnum):
        """
        Column indices of the table
        """

        KEY = 0
        VALUE = 1

    def add_key_value_pair(self, key: str, value: str) -> None:
        """
        Add a key-value pair to the table.
        """
        row_count = self.rowCount()
        key_item = QtWidgets.QTableWidgetItem(key)
        key_item.setFlags(
            QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable  # type: ignore
        )
        value_item = QtWidgets.QTableWidgetItem(value)
        value_item.setFlags(
            QtCore.Qt.ItemIsEnabled  # type: ignore
            | QtCore.Qt.ItemIsEditable
            | QtCore.Qt.ItemIsSelectable
        )
        with BlockSignals(self) as blocked_widget:
            blocked_widget.setRowCount(row_count + 1)
            with BlockSignals(blocked_widget.model()):
                blocked_widget.setItem(
                    row_count, KeyValueTable.ColumnIndex.KEY, key_item
                )
                blocked_widget.setItem(
                    row_count, KeyValueTable.ColumnIndex.VALUE, value_item
                )

    def remove_selected(self) -> str:
        """
        Selections are single-row selection.
        Remove that selected row, and return the key string for that removed row.
        """
        selected_ranges = self.selectedRanges()
        assert selected_ranges
        row_index = selected_ranges[0].topRow()
        assert row_index == selected_ranges[0].bottomRow()
        selected_key = self.item(row_index, KeyValueTable.ColumnIndex.KEY).text()
        self.removeRow(row_index)
        return selected_key



#!/usr/bin/env python3

"""
Remote browser page
"""

import typing

from qtpy import QtCore, QtGui, QtWidgets

from cruiz.interop.reciperevisionsparameters import RecipeRevisionsParameters

from .page import Page


class _RecipeRevisionModel(QtCore.QAbstractTableModel):
    def __init__(self) -> None:
        super().__init__()
        self._rrevs: typing.Optional[typing.List[typing.Dict[str, str]]] = None

    def set(self, results: typing.Optional[typing.List[typing.Dict[str, str]]]) -> None:
        """
        Set the results against the model
        """
        self.beginResetModel()
        self._rrevs = results
        self.endResetModel()

    def rowCount(self, parent) -> int:  # type: ignore
        if parent.isValid():
            return 0
        if self._rrevs is None:
            return 0
        return len(self._rrevs)

    def columnCount(self, parent) -> int:  # type: ignore
        # pylint: disable=unused-argument
        if self._rrevs is None:
            return 0
        return 2

    def headerData(self, section, orientation, role):  # type: ignore
        if role == QtCore.Qt.DisplayRole and orientation == QtCore.Qt.Horizontal:
            if section == 0:
                return "Revision"
            if section == 1:
                return "Timestamp"
        return None

    def data(self, index, role) -> typing.Any:  # type: ignore
        if role == QtCore.Qt.DisplayRole:
            assert self._rrevs
            if index.column() == 0:
                return self._rrevs[index.row()]["revision"]
            if index.column() == 1:
                return self._rrevs[index.row()]["time"]
        return None


class RecipeRevisionPage(Page):
    """
    Remote browser page for recipe revisions
    """

    def setup(self, self_ui: typing.Any) -> None:
        """
        Setup the UI for the page
        """
        self._base_setup(self_ui, 1)
        self._current_pkgref: typing.Optional[str] = None
        self._model = _RecipeRevisionModel()
        self._ui.recipe_revisions.setModel(self._model)
        self._ui.recipe_revisions.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeToContents
        )
        self._ui.recipe_revisions.customContextMenuRequested.connect(
            self._on_selected_pkgref_menu
        )

        self._ui.recipe_revisions.doubleClicked.connect(self._on_rrev_dclicked)
        self._ui.rrev_back.clicked.connect(self._on_back)
        self._ui.rrev_cancel.clicked.connect(self.on_cancel)
        self._ui.rrev_refresh.clicked.connect(self._on_refresh)
        self._ui.rrev_pkgref.customContextMenuRequested.connect(self._on_pkgref_menu)

    @property
    def package_reference(self) -> str:
        """
        Get the package reference for the selection on this page
        """
        selection = self._ui.recipe_revisions.selectedIndexes()
        assert 2 == len(selection)  # num columns in row
        rrev = self._model.data(selection[0], QtCore.Qt.DisplayRole)
        return f"{self._previous_pkgref}#{rrev}"

    def _enable_progress(self, enable: bool) -> None:
        self._ui.rrev_progress.setMaximum(0 if enable else 1)
        self._ui.rrev_buttons.setEnabled(not enable)
        self._ui.rrev_cancel.setEnabled(enable)

    def showEvent(self, event: QtGui.QShowEvent) -> None:
        # pylint: disable=unused-argument
        self._ui.rrev_pkgref.setText(self._previous_pkgref)
        self._compute()

    def _complete(self, results: typing.Any, exception: typing.Any) -> None:
        self._enable_progress(False)
        if exception:
            self._ui.rrev_groupbox.setTitle("Error finding recipe revision")
            self._log_details.stderr(str(exception))
            return
        if results is not None:
            self._model.set(results)
            self._ui.rrev_groupbox.setTitle(f"{len(results)} recipe revisions found")
            self._ui.rrev_groupbox.setEnabled(True)
            self._current_pkgref = self._previous_pkgref

    def _on_back(self) -> None:
        self._open_previous_page()

    def _on_rrev_dclicked(self, index: QtCore.QModelIndex) -> None:
        # pylint: disable=unused-argument
        self._open_next_page()

    def on_cancel(self) -> None:
        """
        Called when the user cancels the operation.
        """
        self._context.cancel()
        self._enable_progress(False)

    def _on_refresh(self) -> None:
        self._current_pkgref = None
        self._compute()

    def _compute(self) -> None:
        pkgref = self._previous_pkgref
        if self._current_pkgref != pkgref:
            self._ui.pkgref_groupbox.setTitle("0 recipe revisions found")
            self._model.set(None)
            self._enable_progress(True)
            self._ui.rrev_groupbox.setEnabled(False)
            params = RecipeRevisionsParameters(
                reference=pkgref,
                remote_name=self._ui.remote.currentText(),
            )
            self._context.get_package_details(params, self._log_details, self._complete)

    def _on_copy_pkgref_to_clip(self) -> None:
        QtWidgets.QApplication.clipboard().setText(self._ui.rrev_pkgref.text())



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

import multiprocessing

from cruiz.interop.commandparameters import CommandParameters
from .utils import worker
from .utils.message import Message, Success


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run 'conan imports'
    """
    with worker.ConanWorker(queue, params) as api:
        assert params.recipe_path
        assert params.cwd
        args = {
            "path": str(params.recipe_path),
        }
        if params.install_folder:
            args["info_folder"] = str(params.install_folder)
        if params.imports_folder:
            args["dest"] = str(params.imports_folder)
        result = api.imports(**args)

        queue.put(Success(result))



#!/usr/bin/env python3

"""
Child process commands
"""

from __future__ import annotations

from io import StringIO
import multiprocessing

from cruiz.interop.commandparameters import CommandParameters
from .utils import worker
from .utils.message import Message, Success, Stdout


def invoke(queue: multiprocessing.Queue[Message], params: CommandParameters) -> None:
    """
    Run 'conan config install'
    """
    with worker.ConanWorker(queue, params) as api:
        args = {}
        if "gitBranch" in params.named_arguments:
            args["args"] = f"-b {params.named_arguments['gitBranch']}"
        if "sourceFolder" in params.named_arguments:
            args["source_folder"] = params.named_arguments["sourceFolder"]
        if "targetFolder" in params.named_arguments:
            args["target_folder"] = params.named_arguments["targetFolder"]

        result = api.config_install(
            params.named_arguments["pathOrUrl"],
            False,  # TODO do we want verify SSL to be false?
            **args,
        )
        message = StringIO()
        message.write(
            "Configuration installed from " f'{params.named_arguments["pathOrUrl"]}'
        )
        if "args" in args:
            message.write(f" with arguments {args['args']} ")
        if "source_folder" in args:
            message.write(f" from source folder {args['source_folder']} ")
        if "target_folder" in args:
            message.write(f" to target folder {args['target_folder']} ")
        queue.put(Stdout(message.getvalue()))
        queue.put(Success(result))



#!/usr/bin/env python3

"""
QLineEdit that can capture keyboard shortcuts
"""

from qtpy import QtCore, QtGui, QtWidgets


class ShortcutLineEdit(QtWidgets.QLineEdit):
    """
    A QLineEdit specifically for capturing keyboard shortcuts
    """

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        assert event.type() == QtCore.QEvent.KeyPress
        key = event.key()
        if key in (  # type: ignore[comparison-overlap]
            QtCore.Qt.Key_unknown,
            QtCore.Qt.Key_Control,
            QtCore.Qt.Key_Shift,
            QtCore.Qt.Key_Alt,
            QtCore.Qt.Key_Meta,
        ):
            return
        if key == QtCore.Qt.Key_Backspace:  # type: ignore[comparison-overlap]
            super().keyPressEvent(event)
            return
        modifiers = event.modifiers()
        if modifiers & QtCore.Qt.ControlModifier:
            key += QtCore.Qt.CTRL  # type: ignore[operator]
        if modifiers & QtCore.Qt.ShiftModifier:
            key += QtCore.Qt.SHIFT  # type: ignore[operator]
        if modifiers & QtCore.Qt.AltModifier:
            key += QtCore.Qt.ALT  # type: ignore[operator]
        if modifiers & QtCore.Qt.MetaModifier:
            key += QtCore.Qt.META  # type: ignore[operator]
        sequence = QtGui.QKeySequence(key)
        self.setText(sequence.toString(QtGui.QKeySequence.PortableText))
        event.accept()



#!/usr/bin/env python3

"""
Recipe command toolbar
"""

from io import StringIO
import os
import typing

from qtpy import QtCore, QtGui, QtWidgets, PYSIDE6
from cruiz.settings.managers.generalpreferences import GeneralSettingsReader

from cruiz.settings.managers.recipe import RecipeSettings, RecipeSettingsReader
from cruiz.settings.managers.shortcuts import ShortcutSettingsReader
from cruiz.settings.managers.compilercachepreferences import CompilerCacheSettingsReader

import cruiz.workers.install

from cruiz.interop.commandparameters import CommandParameters

from cruiz.constants import BuildFeatureConstants, CompilerCacheTypes
from cruiz.environ import EnvironSaver

from cruiz.exceptions import RecipeInspectionError

if PYSIDE6:
    QAction = QtGui.QAction
    QActionGroup = QtGui.QActionGroup
else:
    QAction = QtWidgets.QAction
    QActionGroup = QtWidgets.QActionGroup


class RecipeCommandToolbar(QtWidgets.QToolBar):
    """
    QToolBar representing the recipe commands
    """

    command_started = QtCore.Signal()
    command_ended = QtCore.Signal()

    def __init__(self, parent: QtWidgets.QWidget) -> None:
        super().__init__(parent)
        recipe_ui = parent._ui
        self.command_started.connect(self._command_started)
        self.command_ended.connect(self._command_ended)
        self._idle_group = QActionGroup(self)
        self._idle_group.setEnabled(True)
        self._cancel_command_group = QActionGroup(self)
        self._cancel_command_group.setEnabled(False)
        self._add_toolbutton(
            [recipe_ui.actionCreateCommand, recipe_ui.actionCreateUpdateCommand]
        )
        self.addSeparator()
        self._add_toolbutton(
            [recipe_ui.actionInstallCommand, recipe_ui.actionInstallUpdateCommand]
        )
        self._add_toolbutton([recipe_ui.actionImportsCommand])
        self._add_toolbutton([recipe_ui.actionSourceCommand])
        self._add_toolbutton([recipe_ui.actionBuildCommand])
        self._add_toolbutton([recipe_ui.actionPackageCommand])
        self._add_toolbutton([recipe_ui.actionExportPackageCommand])
        self._add_toolbutton([recipe_ui.actionTestCommand])
        self.addSeparator()
        self._add_toolbutton([recipe_ui.actionCancelCommand], for_cancel_group=True)
        self.addSeparator()
        self._add_toolbutton([recipe_ui.actionRemovePackageCommand])
        self.addSeparator()
        self._add_toolbutton(
            [
                recipe_ui.actionCMakeBuildToolCommand,
                recipe_ui.actionCMakeBuildToolVerboseCommand,
                recipe_ui.actionCMakeRemoveCacheCommand,
            ]
        )

    @property
    def _recipe_widget(self) -> QtCore.QObject:
        return self.parent()

    def _add_toolbutton(
        self, actions: typing.List[QAction], for_cancel_group: bool = False
    ) -> None:
        assert actions
        button = QtWidgets.QToolButton()
        button.addActions(actions)
        button.setDefaultAction(actions[0])
        self.addWidget(button)
        if for_cancel_group:
            for action in actions:
                self._cancel_command_group.addAction(action)
        else:
            for action in actions:
                self._idle_group.addAction(action)

    def configure_actions(self) -> None:
        """
        Configure the QActions in the toolbar
        """

        def _configure(
            action: QAction, trigger_slot: typing.Callable[[], None]
        ) -> None:
            action.setShortcutVisibleInContextMenu(True)
            action.setShortcutContext(QtCore.Qt.WindowShortcut)
            action.triggered.connect(trigger_slot)

        # shortcuts themselves are set elsewhere, as they can be dynamic
        # through the lifetime of the application
        recipe_ui = self.parent()._ui
        _configure(recipe_ui.actionCreateCommand, self._conan_create)
        _configure(recipe_ui.actionCreateUpdateCommand, self._conan_create_update)
        _configure(recipe_ui.actionInstallCommand, self._conan_install)
        _configure(recipe_ui.actionInstallUpdateCommand, self._conan_install_update)
        _configure(recipe_ui.actionImportsCommand, self._conan_imports)
        _configure(recipe_ui.actionSourceCommand, self._conan_source)
        _configure(recipe_ui.actionBuildCommand, self._conan_build)
        _configure(recipe_ui.actionPackageCommand, self._conan_package)
        _configure(recipe_ui.actionExportPackageCommand, self._conan_export_package)
        _configure(recipe_ui.actionTestCommand, self._conan_test)
        _configure(recipe_ui.actionRemovePackageCommand, self._conan_remove)
        _configure(recipe_ui.actionCancelCommand, self._cancel_command)
        _configure(recipe_ui.actionCMakeBuildToolCommand, self._cmake_build)
        _configure(
            recipe_ui.actionCMakeBuildToolVerboseCommand, self._cmake_build_verbose
        )
        _configure(recipe_ui.actionCMakeRemoveCacheCommand, self._cmake_remove_cache)

    def disable_all_actions(self) -> None:
        """
        Disable everything in the case of recipe loading errors
        """
        self._idle_group.setEnabled(False)
        self._cancel_command_group.setEnabled(False)

    def refresh_action_shortcuts_and_tooltips(
        self, recipe_attributes: typing.Dict[str, typing.Optional[str]]
    ) -> None:
        """
        Refresh all command action shortcuts and tooltips
        """
        with ShortcutSettingsReader() as settings:
            conan_create = settings.conan_create.resolve()
            conan_create_updates = settings.conan_create_updates.resolve()
            conan_imports = settings.conan_imports.resolve()
            conan_install = settings.conan_install.resolve()
            conan_install_updates = settings.conan_install_updates.resolve()
            conan_source = settings.conan_source.resolve()
            conan_build = settings.conan_build.resolve()
            conan_package = settings.conan_package.resolve()
            conan_exportpkg = settings.conan_export_package.resolve()
            conan_test = settings.conan_test_package.resolve()
            conan_remove = settings.conan_remove_package.resolve()
            cancel = settings.cancel.resolve()
            cmake_build_tool = settings.cmake_build_tool.resolve()
            cmake_build_tool_verbose = settings.cmake_build_tool_verbose.resolve()
            remove_cmakecache = settings.delete_cmake_cache.resolve()

        def _configure(
            action: QAction, shortcut: str, params: CommandParameters
        ) -> None:
            action.setShortcut(QtGui.QKeySequence(shortcut))
            action.setToolTip(self._generate_command_tooltip(params))

        recipe_ui = self.parent()._ui
        _configure(
            recipe_ui.actionCreateCommand,
            conan_create,
            self._make_conan_create_params(recipe_attributes, None),
        )
        _configure(
            recipe_ui.actionCreateUpdateCommand,
            conan_create_updates,
            self._make_conan_create_params(recipe_attributes, ["-u"]),
        )
        _configure(
            recipe_ui.actionInstallCommand,
            conan_install,
            self._make_conan_install_params(recipe_attributes, None),
        )
        _configure(
            recipe_ui.actionInstallUpdateCommand,
            conan_install_updates,
            self._make_conan_install_params(recipe_attributes, ["-u"]),
        )
        _configure(
            recipe_ui.actionImportsCommand,
            conan_imports,
            self._make_conan_imports_params(recipe_attributes),
        )
        _configure(
            recipe_ui.actionSourceCommand,
            conan_source,
            self._make_conan_source_params(recipe_attributes),
        )
        _configure(
            recipe_ui.actionBuildCommand,
            conan_build,
            self._make_conan_build_params(recipe_attributes),
        )
        _configure(
            recipe_ui.actionPackageCommand,
            conan_package,
            self._make_conan_package_params(recipe_attributes),
        )
        _configure(
            recipe_ui.actionExportPackageCommand,
            conan_exportpkg,
            self._make_conan_export_package_params(recipe_attributes),
        )
        _configure(
            recipe_ui.actionTestCommand,
            conan_test,
            self._make_conan_test_package_params(recipe_attributes),
        )
        _configure(
            recipe_ui.actionRemovePackageCommand,
            conan_remove,
            self._make_conan_remove_package_params(recipe_attributes),
        )
        recipe_ui.actionCancelCommand.setShortcut(QtGui.QKeySequence(cancel))
        recipe_ui.actionCancelCommand.setToolTip("Cancel the currently running command")
        recipe_ui.actionCMakeBuildToolCommand.setShortcut(
            QtGui.QKeySequence(cmake_build_tool)
        )
        recipe_ui.actionCMakeBuildToolCommand.setToolTip("CMake build")
        recipe_ui.actionCMakeBuildToolVerboseCommand.setShortcut(
            QtGui.QKeySequence(cmake_build_tool_verbose)
        )
        recipe_ui.actionCMakeBuildToolVerboseCommand.setToolTip("CMake verbose build")
        recipe_ui.actionCMakeRemoveCacheCommand.setShortcut(
            QtGui.QKeySequence(remove_cmakecache)
        )
        recipe_ui.actionCMakeRemoveCacheCommand.setToolTip("Remove CMake cache")

    def _command_started(self) -> None:
        self._idle_group.setEnabled(False)
        self._cancel_command_group.setEnabled(True)

    def _command_ended(self) -> None:
        self._idle_group.setEnabled(True)
        self._cancel_command_group.setEnabled(False)

    def _append_build_features(
        self, params: CommandParameters, settings: RecipeSettings
    ) -> None:
        if settings.cmake_find_debug.resolve():
            params.set_build_feature(BuildFeatureConstants.CMAKEFINDDEBUGMODE, "ON")
        if settings.cmake_verbose.resolve():
            params.set_build_feature(BuildFeatureConstants.CMAKEVERBOSEMODE, "ON")
        compilercache_autotools_config = (
            settings.compilercache_autotools_configuration.resolve()
        )

        def _find_executable(suggested_dir: str, executable_name: str) -> str:
            with EnvironSaver():
                if suggested_dir:
                    os.environ["PATH"] = (
                        QtCore.QDir.toNativeSeparators(suggested_dir)
                        + os.pathsep
                        + os.environ.get("PATH", "")
                    )
                exe_path = QtCore.QStandardPaths.findExecutable(executable_name)
                return exe_path

        compiler_cache = settings.compiler_cache.resolve()
        if compiler_cache == CompilerCacheTypes.CCACHE.value:
            with CompilerCacheSettingsReader() as settings_compilercache:
                ccache_dir = settings_compilercache.ccache_bin_directory.resolve()
            executable = _find_executable(ccache_dir, "ccache")
            if executable:
                params.set_build_feature(
                    BuildFeatureConstants.CCACHEEXECUTABLE, executable
                )
                if CompilerCacheTypes.CCACHE.name in compilercache_autotools_config:
                    params.set_build_feature(
                        BuildFeatureConstants.CCACHEAUTOTOOLSCONFIGARGS,
                        compilercache_autotools_config[CompilerCacheTypes.CCACHE.name],
                    )
        elif compiler_cache == CompilerCacheTypes.SCCACHE.value:
            with CompilerCacheSettingsReader() as settings_compilercache:
                sccache_dir = settings_compilercache.sccache_bin_directory.resolve()
            executable = _find_executable(sccache_dir, "sccache")
            if executable:
                params.set_build_feature(
                    BuildFeatureConstants.SCCACHEEXECUTABLE, executable
                )
                if CompilerCacheTypes.SCCACHE.name in compilercache_autotools_config:
                    params.set_build_feature(
                        BuildFeatureConstants.SCCACHEAUTOTOOLSCONFIGARGS,
                        compilercache_autotools_config[CompilerCacheTypes.SCCACHE.name],
                    )
        elif compiler_cache == CompilerCacheTypes.BUILDCACHE.value:
            with CompilerCacheSettingsReader() as settings_compilercache:
                buildcache_dir = (
                    settings_compilercache.buildcache_bin_directory.resolve()
                )
            executable = _find_executable(buildcache_dir, "buildcache")
            if executable:
                params.set_build_feature(
                    BuildFeatureConstants.BUILDCACHEEXECUTABLE, executable
                )
                if CompilerCacheTypes.BUILDCACHE.name in compilercache_autotools_config:
                    params.set_build_feature(
                        BuildFeatureConstants.BUILDCACHEAUTOTOOLSCONFIGARGS,
                        compilercache_autotools_config[
                            CompilerCacheTypes.BUILDCACHE.name
                        ],
                    )

    def _append_general_prefs(self, params: CommandParameters) -> None:
        with GeneralSettingsReader() as settings:
            params.time_commands = settings.enable_command_timing.resolve()

    def _make_common_params(
        self,
        verb: str,
        worker: typing.Union[
            typing.Callable[[typing.Any, typing.Any], None],
            typing.Callable[[typing.Any, typing.Any, typing.Any], None],
        ],
        recipe_attributes: typing.Dict[str, typing.Optional[str]],
        with_recipe_path: bool = True,
        with_pkgref: bool = False,
        with_cwd: bool = False,
        with_profile: bool = False,
        with_install_folder: bool = False,
        with_imports_folder: bool = False,
        with_source_folder: bool = False,
        fudge_source_folder: bool = False,
        with_build_folder: bool = False,
        with_package_folder: bool = False,
        with_test_folder: bool = False,
        with_explicit_name: bool = False,
        with_force: bool = False,
        with_exclusive_package_folder: bool = False,
        with_options: bool = False,
    ) -> CommandParameters:
        recipe_widget = self._recipe_widget
        recipe = recipe_widget.recipe
        params = CommandParameters(verb, worker)
        if with_recipe_path:
            params.recipe_path = recipe.path
        if with_explicit_name:
            params.name = recipe_attributes["name"]
        if with_pkgref:
            params.version = recipe.version or recipe_attributes["version"]
            params.user = recipe.user
            params.channel = recipe.channel
            params.make_package_reference()  # only needed for the exported string
        if with_force:
            params.force = True
        self._append_general_prefs(params)
        with RecipeSettingsReader.from_recipe(recipe) as settings:
            num_cores = settings.num_cpu_cores
            if num_cores.value is not None:
                params.added_environment.update(
                    {"CONAN_CPU_COUNT": str(num_cores.value)}
                )
            if with_options:
                for key, value in settings.options.resolve().items():
                    assert recipe_attributes["name"]
                    params.add_option(
                        recipe_attributes["name"], key, value
                    )  # TODO: is this the most efficient algorithm?
            self._append_build_features(params, settings)
            if with_exclusive_package_folder:
                # export-pkg requires
                # - if package_folder is present, do not specify source or build folder
                # - if package_folder is NOT present, you might still need the fudge of
                #   the source folder
                if settings.local_workflow_package_folder.resolve() is None:
                    with_source_folder = True
                    with_build_folder = True
                else:
                    with_source_folder = False
                    with_build_folder = False
                    fudge_source_folder = False
            if with_profile:
                params.profile = settings.profile.resolve()
            if with_cwd:
                workflow_cwd = settings.local_workflow_cwd.resolve()
                common_subdir = settings.local_workflow_common_subdir.resolve()
                params.cwd = recipe_widget.get_working_dir(workflow_cwd, common_subdir)
            if with_install_folder:
                install_folder = settings.local_workflow_install_folder.resolve()
                params.install_folder = recipe_widget.resolve_expression(install_folder)
            if with_imports_folder:
                imports_folder = settings.local_workflow_imports_folder.resolve()
                params.imports_folder = recipe_widget.resolve_expression(imports_folder)
            if with_source_folder:
                source_folder = settings.local_workflow_source_folder.resolve()
                params.source_folder = recipe_widget.resolve_expression(source_folder)
            if fudge_source_folder:
                source_folder = settings.local_workflow_source_folder.resolve()
                if (
                    source_folder is None
                    and not recipe_widget.cwd_is_relative_to_recipe(workflow_cwd)
                ):
                    # fudge due to the default source_folder changing between
                    # 'conan source' and 'conan build'
                    source_folder = "."
                params.source_folder = recipe_widget.resolve_expression(source_folder)
            if with_build_folder:
                build_folder = settings.local_workflow_build_folder.resolve()
                params.build_folder = recipe_widget.resolve_expression(build_folder)
            if with_package_folder:
                package_folder = settings.local_workflow_package_folder.resolve()
                params.package_folder = recipe_widget.resolve_expression(package_folder)
            if with_test_folder:
                test_folder = settings.local_workflow_test_folder.resolve()
                params.test_folder = recipe_widget.resolve_expression(test_folder)
        return params

    def _make_conan_create_params(
        self,
        recipe_attributes: typing.Dict[str, typing.Optional[str]],
        args: typing.Optional[typing.List[str]],
    ) -> CommandParameters:
        params = self._make_common_params(
            "create",
            cruiz.workers.create.invoke,
            recipe_attributes,
            with_pkgref=True,
            with_profile=True,
            with_options=True,
        )
        if args:
            params.arguments.extend(args)
        return params

    def _make_conan_install_params(
        self,
        recipe_attributes: typing.Dict[str, typing.Optional[str]],
        args: typing.Optional[typing.List[str]],
    ) -> CommandParameters:
        params = self._make_common_params(
            "install",
            cruiz.workers.install.invoke,
            recipe_attributes,
            with_pkgref=True,
            with_cwd=True,
            with_profile=True,
            with_install_folder=True,
            with_options=True,
        )
        if args:
            params.arguments.extend(args)
        return params

    def _make_conan_imports_params(
        self, recipe_attributes: typing.Dict[str, typing.Optional[str]]
    ) -> CommandParameters:
        return self._make_common_params(
            "imports",
            cruiz.workers.imports.invoke,
            recipe_attributes,
            with_cwd=True,
            with_install_folder=True,
            with_imports_folder=True,
        )

    def _make_conan_source_params(
        self, recipe_attributes: typing.Dict[str, typing.Optional[str]]
    ) -> CommandParameters:
        return self._make_common_params(
            "source",
            cruiz.workers.source.invoke,
            recipe_attributes,
            with_cwd=True,
            with_install_folder=True,
            with_source_folder=True,
        )

    def _make_conan_build_params(
        self, recipe_attributes: typing.Dict[str, typing.Optional[str]]
    ) -> CommandParameters:
        return self._make_common_params(
            "build",
            cruiz.workers.build.invoke,
            recipe_attributes,
            with_cwd=True,
            with_install_folder=True,
            with_source_folder=True,
            fudge_source_folder=True,
            with_build_folder=True,
            with_package_folder=True,
        )

    def _make_conan_package_params(
        self, recipe_attributes: typing.Dict[str, typing.Optional[str]]
    ) -> CommandParameters:
        return self._make_common_params(
            "package",
            cruiz.workers.package.invoke,
            recipe_attributes,
            with_cwd=True,
            with_install_folder=True,
            with_source_folder=True,
            fudge_source_folder=True,
            with_build_folder=True,
            with_package_folder=True,
        )

    def _make_conan_export_package_params(
        self, recipe_attributes: typing.Dict[str, typing.Optional[str]]
    ) -> CommandParameters:
        return self._make_common_params(
            "export-pkg",
            cruiz.workers.exportpackage.invoke,
            recipe_attributes,
            with_cwd=True,
            with_pkgref=True,
            with_install_folder=True,
            with_source_folder=True,
            fudge_source_folder=True,
            with_build_folder=True,
            with_package_folder=True,
            with_exclusive_package_folder=True,
            with_explicit_name=True,
            with_force=True,
            with_options=True,
        )

    def _make_conan_test_package_params(
        self, recipe_attributes: typing.Dict[str, typing.Optional[str]]
    ) -> CommandParameters:
        params = self._make_common_params(
            "test",
            cruiz.workers.testpackage.invoke,
            recipe_attributes,
            with_recipe_path=False,
            with_cwd=True,
            with_profile=True,
            with_test_folder=True,
            with_explicit_name=True,
            with_pkgref=True,
            with_options=True,
        )
        recipe_widget = self._recipe_widget
        recipe = recipe_widget.recipe
        params.recipe_path = recipe.path.parent / "test_package"
        return params

    def _make_conan_remove_package_params(
        self, recipe_attributes: typing.Dict[str, typing.Optional[str]]
    ) -> CommandParameters:
        return self._make_common_params(
            "remove",
            cruiz.workers.removepackage.invoke,
            recipe_attributes,
            with_explicit_name=True,
            with_pkgref=True,
            with_force=True,
            with_recipe_path=False,
        )

    def _conan_create_common(
        self, args: typing.Optional[typing.List[str]] = None
    ) -> None:
        # note that this is done early, to force a focus change in any
        # configuration windows, forcing serialising changes before commands kick off
        recipe_widget = self._recipe_widget
        try:
            recipe_attributes = recipe_widget.get_recipe_attributes()
        except RecipeInspectionError:
            return
        self.command_started.emit()
        recipe_widget.recipe.context.conancommand(
            self._make_conan_create_params(recipe_attributes, args),
            recipe_widget.log_details,
            self,
        )

    def _conan_create(self) -> None:
        self._conan_create_common()

    def _conan_create_update(self) -> None:
        self._conan_create_common(args=["--update"])

    def _conan_install_common(
        self, args: typing.Optional[typing.List[str]] = None
    ) -> None:
        recipe_widget = self._recipe_widget
        try:
            recipe_attributes = recipe_widget.get_recipe_attributes()
        except RecipeInspectionError:
            return
        self.command_started.emit()
        recipe_widget.recipe.context.conancommand(
            self._make_conan_install_params(recipe_attributes, args),
            recipe_widget.log_details,
            self,
        )

    def _conan_install(self) -> None:
        self._conan_install_common()

    def _conan_install_update(self) -> None:
        self._conan_install_common(args=["--update"])

    def _conan_imports(self) -> None:
        recipe_widget = self._recipe_widget
        try:
            recipe_attributes = recipe_widget.get_recipe_attributes()
        except RecipeInspectionError:
            return
        self.command_started.emit()
        recipe_widget.recipe.context.conancommand(
            self._make_conan_imports_params(recipe_attributes),
            recipe_widget.log_details,
            self,
        )

    def _conan_source(self) -> None:
        recipe_widget = self._recipe_widget
        try:
            recipe_attributes = recipe_widget.get_recipe_attributes()
        except RecipeInspectionError:
            return
        self.command_started.emit()
        recipe_widget.recipe.context.conancommand(
            self._make_conan_source_params(recipe_attributes),
            recipe_widget.log_details,
            self,
        )

    def _conan_build(self) -> None:
        recipe_widget = self._recipe_widget
        try:
            recipe_attributes = recipe_widget.get_recipe_attributes()
        except RecipeInspectionError:
            return
        self.command_started.emit()
        recipe_widget.recipe.context.conancommand(
            self._make_conan_build_params(recipe_attributes),
            recipe_widget.log_details,
            self,
        )

    def _conan_package(self) -> None:
        recipe_widget = self._recipe_widget
        try:
            recipe_attributes = recipe_widget.get_recipe_attributes()
        except RecipeInspectionError:
            return
        self.command_started.emit()
        recipe_widget.recipe.context.conancommand(
            self._make_conan_package_params(recipe_attributes),
            recipe_widget.log_details,
            self,
        )

    def _conan_export_package(self) -> None:
        recipe_widget = self._recipe_widget
        try:
            recipe_attributes = recipe_widget.get_recipe_attributes()
        except RecipeInspectionError:
            return
        self.command_started.emit()
        recipe_widget.recipe.context.conancommand(
            self._make_conan_export_package_params(recipe_attributes),
            recipe_widget.log_details,
            self,
        )

    def _conan_test(self) -> None:
        recipe_widget = self._recipe_widget
        try:
            recipe_attributes = recipe_widget.get_recipe_attributes()
        except RecipeInspectionError:
            return
        self.command_started.emit()
        recipe_widget.recipe.context.conancommand(
            self._make_conan_test_package_params(recipe_attributes),
            recipe_widget.log_details,
            self,
        )

    def _cancel_command(self) -> None:
        recipe_widget = self._recipe_widget
        recipe = recipe_widget.recipe
        recipe.context.cancel()

    def _conan_remove(self) -> None:
        recipe_widget = self._recipe_widget
        try:
            recipe_attributes = recipe_widget.get_recipe_attributes()
        except RecipeInspectionError:
            return
        self.command_started.emit()
        recipe_widget.recipe.context.conancommand(
            self._make_conan_remove_package_params(recipe_attributes),
            recipe_widget.log_details,
            self,
        )

    def _cmake_build_common(
        self, args: typing.Optional[typing.List[str]] = None
    ) -> None:
        # TODO review
        recipe_widget = self._recipe_widget
        recipe = recipe_widget.recipe
        log_details = recipe_widget.log_details
        params = CommandParameters(
            "cmakebuild", cruiz.workers.cmakebuildtool.invoke
        )  # TODO: verb is wrong
        with RecipeSettingsReader.from_recipe(recipe) as settings:
            num_cores = settings.num_cpu_cores
            if num_cores.value is not None:
                params.added_environment.update(
                    {"CONAN_CPU_COUNT": str(num_cores.value)}
                )
            workflow_cwd = settings.local_workflow_cwd.resolve()
            common_subdir = settings.local_workflow_common_subdir.resolve()
            params.cwd = recipe_widget.get_working_dir(workflow_cwd, common_subdir)
            build_folder = settings.local_workflow_build_folder.resolve()
            params.build_folder = recipe_widget.resolve_expression(build_folder)
        if args:
            params.arguments.extend(args)
        self.command_started.emit()
        recipe.context.cmakebuildcommand(params, log_details, self)

    def _cmake_build(self) -> None:
        self._cmake_build_common()

    def _cmake_build_verbose(self) -> None:
        self._cmake_build_common(args=["verbose"])

    def _cmake_remove_cache(self) -> None:
        # TODO: review
        recipe_widget = self._recipe_widget
        recipe = recipe_widget.recipe
        log_details = recipe_widget.log_details
        params = CommandParameters(
            "cmakeremovecache", cruiz.workers.deletecmakecache.invoke
        )  # TODO: verb is wrong
        with RecipeSettingsReader.from_recipe(recipe) as settings:
            workflow_cwd = settings.local_workflow_cwd.resolve()
            common_subdir = settings.local_workflow_common_subdir.resolve()
            params.cwd = recipe_widget.get_working_dir(workflow_cwd, common_subdir)
            build_folder = settings.local_workflow_build_folder.resolve()
            params.build_folder = recipe_widget.resolve_expression(build_folder)
        self.command_started.emit()
        recipe.context.cmakebuildcommand(params, log_details, self)

    def _generate_command_tooltip(self, params: CommandParameters) -> str:
        tooltip = StringIO()
        tooltip.write(f"<p style='white-space:pre'><b>{params}</b></p>\n")
        if params.cwd:
            tooltip.write("<p style='white-space:pre'>Working directory<br>")
            tooltip.write(f" <em>{params.cwd}</em></p>\n")
        if params.added_environment:
            tooltip.write("<p style='white-space:pre'>Environment<br>")
            for key, value in params.added_environment.items():
                tooltip.write(f" {key}={value}<br>")
            tooltip.write("</p>")
        return tooltip.getvalue()



#!/usr/bin/env python3

"""
Remote browser page
"""

import typing

from qtpy import QtCore, QtGui, QtWidgets, PYSIDE2

if PYSIDE2:
    QAction = QtWidgets.QAction
else:
    QAction = QtGui.QAction

from cruiz.commands.context import ConanContext
from cruiz.commands.logdetails import LogDetails


class Page(QtWidgets.QWidget):
    """
    Base class for all pages shown in the remote browser
    """

    def _base_setup(self, self_ui: typing.Any, index: int) -> None:
        self._ui = self_ui
        self.page_index = index

    @property
    def _log_details(self) -> LogDetails:
        stacked_widget = self.parent()
        dock_contents = stacked_widget.parent()
        remote_browser = dock_contents.parent()
        return remote_browser._log_details

    @property
    def _context(self) -> ConanContext:
        stacked_widget = self.parent()
        dock_contents = stacked_widget.parent()
        remote_browser = dock_contents.parent()
        return remote_browser._context

    @property
    def _revisions_enabled(self) -> bool:
        stacked_widget = self.parent()
        pkgref_page = stacked_widget.widget(0)
        return pkgref_page._revs_enabled

    def _on_pkgref_menu(self, position: QtCore.QPoint) -> None:
        menu = QtWidgets.QMenu(self)
        copy_action = QAction("Copy to clipboard", self)
        copy_action.triggered.connect(self._on_copy_pkgref_to_clip)
        menu.addAction(copy_action)
        menu.exec_(self.sender().mapToGlobal(position))

    def _on_selected_pkgref_menu(self, position: QtCore.QPoint) -> None:
        menu = QtWidgets.QMenu(self)
        copy_action = QAction("Copy to clipboard", self)
        copy_action.triggered.connect(self._on_copy_selected_pkgref_to_clip)
        menu.addAction(copy_action)
        menu.exec_(self.sender().mapToGlobal(position))

    def _on_copy_selected_pkgref_to_clip(self) -> None:
        QtWidgets.QApplication.clipboard().setText(self.package_reference)

    def _open_start(self) -> None:
        self.parent().setCurrentIndex(0)
        self.parent().widget(0).invalidate()

    def _open_previous_page(self) -> None:
        self.parent().setCurrentIndex(self.page_index - 1)

    def _open_next_page(self) -> None:
        self.parent().setCurrentIndex(self.page_index + 1)

    @property
    def _previous_page(self) -> QtWidgets.QWidget:
        if self._revisions_enabled:
            return self.parent().widget(self.page_index - 1)
        return self.parent().widget(self.page_index - 2)

    @property
    def _previous_pkgref(self) -> str:
        return self._previous_page.package_reference
